from dataclasses import dataclass
from typing import Generic, TypeVar

from PySide6.QtGui import Qt
from PySide6.QtWidgets import QListWidget, QListWidgetItem

_T = TypeVar("_T")


@dataclass
class ListItem(Generic[_T]):
    display: str
    value: _T


class QList(QListWidget, Generic[_T]):
    def add(self, item: ListItem[_T]) -> None:
        list_item = QListWidgetItem(item.display)
        list_item.setData(Qt.ItemDataRole.UserRole, item)
        self.addItem(list_item)

    def add_all(self, items: list[ListItem[_T]]) -> None:
        self.clear()
        for item in items:
            self.add(item)

    def values(self) -> list[_T]:
        return [self.item(i).data(Qt.ItemDataRole.UserRole).value for i in range(self.count())]

    def get_selected_item(self) -> _T | None:
        current_row = self.currentRow()
        if current_row >= 0:
            return self.item(current_row).data(Qt.ItemDataRole.UserRole).value
        return None

    def delete_selected_item(self) -> None:
        current_row = self.currentRow()
        if current_row >= 0:
            self.takeItem(current_row)

    def sort_by_value(self):
        items_display = self._get_items_display()
        sorted_items = sorted(  # if the user passes values that cannot be sorted, it will fail
            items_display, key=lambda x: x.value  # pyright: ignore [reportArgumentType, reportCallIssue]
        )

        self.add_all(sorted_items)

    def unique(self):
        items_display = self._get_items_display()
        unique_data_dict = {}
        for item in items_display:
            unique_data_dict[item.value] = item

        self.add_all(list(unique_data_dict.values()))

    def _get_items_display(self) -> list[ListItem[_T]]:
        items = [self.item(i) for i in range(self.count())]
        return [item.data(Qt.ItemDataRole.UserRole) for item in items]



from pathlib import Path

from clochette import log
from clochette.persist.configuration.PlatformConfigurationService import PlatformConfigurationService
from clochette.persist import db


class BaseDAO:
    _db_file: Path

    def __init__(self, platform_configuration_service: PlatformConfigurationService):
        data = platform_configuration_service.application_data_path()
        self._db_file = data.joinpath("database.sqlite")
        log.info(f"Database path: {self._db_file}")
        db.init(str(self._db_file), pragmas={"journal_mode": "wal"})

    def create(self, query: str) -> None:
        db.execute_sql(query)



from dataclasses import dataclass

from clochette.framework.qt.Link import Link
from clochette.framework.qt.QComponent import QComponent
from clochette.presentation.settings.calendar.CalendarNavigationModel import CalendarNavigationModel
from clochette.presentation.settings.calendar.QCalendarSettings import QCalendarSettingsView


@dataclass
class QCalendarSettingsComponent(QComponent[QCalendarSettingsView]):
    """Component that wires CalendarNavigationModel to QCalendarSettingsView"""

    _view: QCalendarSettingsView
    _calendar_navigation_model: CalendarNavigationModel

    def __post_init__(self):
        super().__init__(self._view)

        # Wire: CalendarNavigationModel → View methods (Link wraps them in InSignals)
        Link(
            observable=self._calendar_navigation_model.show_calendar_list,
            handler=lambda _: self._view.show_calendar_list(),
            widget=self._view,
        )
        Link(
            observable=self._calendar_navigation_model.add_calendar,
            handler=lambda _: self._view.add_calendar(),
            widget=self._view,
        )
        Link(
            observable=self._calendar_navigation_model.edit_calendar,
            handler=lambda _: self._view.edit_calendar(),
            widget=self._view,
        )



from dataclasses import dataclass

from reactivex import Observable
from reactivex import operators as ops

from clochette import log
from clochette.application.store.GeneralConfigurationStore import GeneralConfigurationStore
from clochette.application.usecase.PersistConfigurationUseCase import PersistConfigurationUseCase
from clochette.domain.entity.delta.SnoozeDelta import SnoozeDelta
from clochette.framework.rx.Scheduler import scheduler


@dataclass
class UpdateSnoozesUseCase:
    """Use case to update snooze configuration in the store."""

    _general_configuration_store: GeneralConfigurationStore
    _persist_configuration_usecase: PersistConfigurationUseCase

    def update_snoozes(self, snoozes: list[SnoozeDelta]) -> Observable[None]:
        """Update snooze deltas and persist to disk."""
        log.info(f"Updating snoozes: {snoozes}")
        self._general_configuration_store.set_snoozes(snoozes)

        # Persist to disk
        return self._persist_configuration_usecase.persist_configuration().pipe(
            ops.map(lambda _: None),
            ops.do_action(on_error=lambda e: log.error(f"Failed to persist snooze configuration", exc_info=e)),
            ops.subscribe_on(scheduler=scheduler),
        )



from configparser import SectionProxy

from clochette.application.service.provider.dto.IProviderConfigurationMapper import IProviderConfigurationMapper
from clochette.persist.configuration.ConfigurationReadUtils import read_str
from clochette.provider.public.configuration.PublicURLSource import PublicURLSource


class PublicURLSourceConfigurationService(IProviderConfigurationMapper[PublicURLSource]):
    _PROVIDER_ID = "public"

    def match(self, section: SectionProxy) -> bool:
        """Check if this mapper can handle the given configuration section"""
        authentication = read_str(section, "authentication", "").lower()
        return authentication == self._PROVIDER_ID

    def read(self, section: SectionProxy) -> PublicURLSource:
        url = read_str(section, "url", "")
        return PublicURLSource(url)

    def write(self, source: PublicURLSource) -> dict[str, str]:
        return {
            "authentication": self._PROVIDER_ID,
            "url": source.url,
        }



from dataclasses import dataclass

from clochette.application.service.provider.dto.ProviderType import ProviderType
from clochette.domain.entity.CalendarID import CalendarID
from clochette.domain.entity.ISourceCalendar import ISourceCalendar
from clochette.framework.qt.QComponent import QComponent
from clochette.presentation.settings.calendar.provider.QAuthenticationSourceView import QAuthenticationSourceView


@dataclass
class QAuthenticationSourceComponent(QComponent[QAuthenticationSourceView]):
    """Component that handles provider-specific authentication logic"""

    _view: QAuthenticationSourceView

    def __post_init__(self):
        super().__init__(self._view)

    def set_values_from_source(self, calendar_id: CalendarID, source: ISourceCalendar) -> None:
        """Set values in the appropriate provider panel based on source type"""
        self._view.set_values_from_source(calendar_id, source)

    def get_source(self) -> ISourceCalendar:
        """Get the calendar source from the selected provider panel"""
        return self._view.get_source()

    def clear(self) -> None:
        """Clear all provider panels"""
        self._view.clear()

    def set_provider_type(self, provider_type: ProviderType) -> None:
        """Set the selected provider type"""
        self._view.set_selected_provider_type(provider_type)



import ssl
import threading
from http.server import HTTPServer
from tempfile import NamedTemporaryFile

from PySide6.QtCore import QFile

from clochette.infrastructure.http_.server.RequestHandlerFactory import HttpRequestHandlerBuilder
from clochette.res.ResUtils import read_file


class SingleRequestHttpsServer:
    HOSTNAME = "localhost"

    _event: threading.Event
    _request_handler_builder: HttpRequestHandlerBuilder
    _http_server: HTTPServer
    _port: int

    def __init__(self, request_handler_builder: HttpRequestHandlerBuilder, port: int = 0):
        self._event = threading.Event()
        self._request_handler_builder = request_handler_builder
        self._port = port

    def _start_server(self) -> None:
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        context.check_hostname = False

        # https://github.com/encode/httpx/discussions/2037#discussioncomment-2006795
        with NamedTemporaryFile(mode="w+b") as cert_file, NamedTemporaryFile(mode="w+b") as key_file:
            cert_file.write(read_file(QFile(":/cert.pem")))
            cert_file.seek(0)

            key_file.write(read_file(QFile(":/key.pem")))
            key_file.seek(0)

            context.load_cert_chain(certfile=cert_file.name, keyfile=key_file.name)

        with HTTPServer(
            (SingleRequestHttpsServer.HOSTNAME, self._port), self._request_handler_builder
        ) as self._http_server:
            self._http_server.socket = context.wrap_socket(self._http_server.socket, server_side=True)
            # handle a single request
            self._event.set()
            self._http_server.serve_forever()

    def start(self) -> None:
        server_thread = threading.Thread(target=self._start_server)
        server_thread.daemon = True
        server_thread.start()
        # wait till the server has been started
        self._event.wait()

    def _stop_server(self) -> None:
        self._http_server.shutdown()

    def stop(self) -> None:
        server_thread = threading.Thread(target=self._stop_server)
        server_thread.daemon = True
        server_thread.start()

    def get_url(self) -> str:
        return f"https://{SingleRequestHttpsServer.HOSTNAME}:{self._http_server.server_port}"



from configparser import SectionProxy

from clochette.application.service.provider.dto.IProviderConfigurationMapper import IProviderConfigurationMapper
from clochette.persist.configuration.ConfigurationReadUtils import read_str
from clochette.provider.basic.configuration.BasicAuthURLSource import BasicAuthURLSource


class BasicAuthURLSourceConfigurationService(IProviderConfigurationMapper[BasicAuthURLSource]):
    _PROVIDER_ID = "basic"

    def match(self, section: SectionProxy) -> bool:
        """Check if this mapper can handle the given configuration section"""
        authentication = read_str(section, "authentication", "").lower()
        return authentication == self._PROVIDER_ID

    def read(self, section: SectionProxy) -> BasicAuthURLSource:
        url = read_str(section, "url", "")
        return BasicAuthURLSource(url)

    def write(self, source: BasicAuthURLSource) -> dict[str, str]:
        return {
            "authentication": self._PROVIDER_ID,
            "url": source.url,
        }



import json

from oauthlib.oauth2 import OAuth2Token


class OAuth2Mapper:

    def to_oauth2_token(self, token: str) -> OAuth2Token:
        return OAuth2Token(json.loads(token))

    def from_oauth2_token(self, token: OAuth2Token) -> str:
        return json.dumps(token)



import configparser
from configparser import SectionProxy
from dataclasses import dataclass
from datetime import timedelta
from pathlib import Path

import isodate
from reactivex import Observable, empty, from_iterable, operators as ops, defer

from clochette import log
from clochette.application.service.configurartion.SourceConfigurationMapperStrategy import (
    SourceConfigurationMapperStrategy,
)
from clochette.domain.entity.delta.AlarmDelta import AlarmDelta
from clochette.domain.entity.CalendarID import CalendarID
from clochette.domain.entity.ISourceCalendar import ISourceCalendar
from clochette.domain.entity.delta.SnoozeDelta import SnoozeDelta
from clochette.domain.entity.configuration.HttpTimeout import HttpTimeout
from clochette.domain.entity.configuration.ThemeConfiguration import ThemeConfiguration
from clochette.domain.entity.configuration.ThemeEnum import ThemeEnum
from clochette.persist.configuration.ConfigurationMigrationService import ConfigurationMigrationService
from clochette.persist.configuration.ConfigurationReadUtils import read_str, read_int
from clochette.persist.configuration.PlatformConfigurationService import PlatformConfigurationService
from clochette.persist.configuration.dto.CalendarConfigurationDTO import CalendarConfigurationDTO
from clochette.persist.configuration.dto.GlobalConfigurationDTO import GlobalConfigurationDTO

_CONFIG_SECTION = "Config"


@dataclass
class ConfigurationReaderService:
    _platform_config: PlatformConfigurationService
    _configuration_migration_service: ConfigurationMigrationService
    _source_configuration_mapper_strategy: SourceConfigurationMapperStrategy

    def read(self) -> Observable[GlobalConfigurationDTO]:
        return defer(lambda _: self._parse_config_file(self._platform_config.application_configuration_file())).pipe(
            ops.do_action(
                on_error=lambda error: log.error(
                    f"Failed to parse config file: {self._platform_config.application_configuration_file()}",
                    exc_info=error,
                )
            ),
        )

    def _parse_config_file(self, configuration_file: Path) -> Observable[GlobalConfigurationDTO]:
        """Parse configuration file and return GlobalConfigurationDTO observable."""
        self._configuration_migration_service.init_application_configuration()
        log.info(f"Start parsing configuration file: {configuration_file}")

        config_parser = configparser.RawConfigParser(
            inline_comment_prefixes=("#", ";"),
        )
        config_parser.read(configuration_file)
        section = config_parser[_CONFIG_SECTION]

        snoozes = self._get_snoozes(section)
        theme = self._get_theme(section)

        return self._get_calendars(config_parser).pipe(
            ops.map(lambda calendars: GlobalConfigurationDTO(snoozes, calendars, theme))
        )

    def _get_theme(self, section: SectionProxy) -> ThemeConfiguration:
        window_icon_theme: ThemeEnum
        systray_icon_theme: ThemeEnum

        window_icon_theme = self._read_theme(section, "theme.icon.window", ThemeEnum.GENERIC)
        systray_icon_theme = self._read_theme(section, "theme.icon.systray", ThemeEnum.GENERIC)

        return ThemeConfiguration(window_icon_theme, systray_icon_theme)

    def _get_calendars(self, config: configparser.RawConfigParser) -> Observable[list[CalendarConfigurationDTO]]:
        sections = [section for section in config.sections() if section != _CONFIG_SECTION]

        return from_iterable(sections).pipe(
            ops.flat_map(lambda calendar_section: self._parse_calendar_section(config, calendar_section)),
            ops.to_list(),
        )

    def _parse_calendar_section(
        self, config: configparser.RawConfigParser, calendar_section: str
    ) -> Observable[CalendarConfigurationDTO]:
        """Parse a single calendar section and return an Observable."""

        log.debug(f"Parsing calendar config section: {calendar_section}")
        section = config[calendar_section]

        return self._read_source(section).pipe(
            ops.map(lambda source: self._build_calendar_dto(section, source, calendar_section)),
            ops.catch(lambda error, _src: self._handle_parse_error(calendar_section, error)),
        )

    def _build_calendar_dto(
        self, section: SectionProxy, source: ISourceCalendar, calendar_section: str
    ) -> CalendarConfigurationDTO:
        """Parse section data and build CalendarConfigurationDTO with the given source."""
        name = read_str(section, "name", "unknown")
        force_alarms = self._get_alarms(section, "force_alarms", [])
        force_alarm_dates = self._get_alarms(section, "force_alarms_dates", [])
        default_alarms = self._get_alarms(section, "default_alarms", [timedelta(minutes=-15)])
        default_alarms_date = self._get_alarms(section, "default_alarms_dates", [])

        download_interval_int = read_int(section, "download.interval.minutes", 5)
        download_interval = timedelta(minutes=download_interval_int)

        missed_reminders_past_window = self._read_iso_duration(
            section, "missed.reminders.past_window", timedelta(hours=-24)
        )

        read_timeout = read_int(section, "timeout.read", 30)
        connection_timeout = read_int(section, "timeout.connection", 30)

        return CalendarConfigurationDTO(
            CalendarID(calendar_section),
            name,
            source,
            force_alarms,
            force_alarm_dates,
            default_alarms,
            default_alarms_date,
            download_interval,
            missed_reminders_past_window,
            HttpTimeout(connection_timeout, read_timeout),
        )

    def _handle_parse_error(self, calendar_section: str, error: Exception) -> Observable[CalendarConfigurationDTO]:
        """Handle errors from parsing a calendar section by logging and returning empty."""
        log.warning(f"Failed to parse section: {calendar_section}", exc_info=error)
        return empty()

    def _read_source(self, section: SectionProxy) -> Observable[ISourceCalendar]:
        return self._source_configuration_mapper_strategy.read(section)

    def _get_snoozes(self, section: SectionProxy) -> list[SnoozeDelta]:
        timedeltas = self._read_duration_array(section, "snooze", [])
        return [SnoozeDelta(td) for td in timedeltas]

    def _get_alarms(self, section: SectionProxy, key_name: str, fallback: list[timedelta]) -> list[AlarmDelta]:
        timedeltas = self._read_duration_array(section, key_name, fallback)
        return [AlarmDelta(td) for td in timedeltas]

    def _read_duration_array(self, section: SectionProxy, key_name: str, fallback: list[timedelta]) -> list[timedelta]:
        values = self._read_array(section, key_name)
        if values is None:
            return fallback

        res = []
        for value in values:
            duration = self._parse_iso_duration(value)
            if duration is not None:
                res.append(duration)

        return res

    def _read_array(self, section: SectionProxy, key_name: str) -> list[str] | None:
        value = section.get(key_name, None)
        if value:
            return list(map(lambda x: x.strip(), value.split(",")))
        else:
            return None

    def _read_iso_duration(self, section: SectionProxy, key_name: str, fallback: timedelta) -> timedelta:
        delta = self._read_iso_duration_strict(section, key_name)
        if delta is None:
            log.warning(f"Failed to parse config item: {key_name}, using fallback: {fallback}")
            return fallback
        else:
            return delta

    def _read_iso_duration_strict(self, section: SectionProxy, key_name: str) -> timedelta | None:
        try:
            value = section.get(key_name, None)
            if value is not None:
                delta = self._parse_iso_duration(value.upper())
                if delta is None:
                    log.warning(f"Failed to parse config item: {key_name}")
                return delta
            else:
                return None
        except Exception:
            log.warning(f"Failed to parse config item: {key_name}", exc_info=True)
            return None

    def _parse_iso_duration(self, duration: str) -> timedelta | None:
        try:
            delta = isodate.parse_duration(duration)
            if isinstance(delta, timedelta):
                return delta
            else:
                return None
        except Exception:
            log.warning(f"Failed to parse config item: {duration}", exc_info=True)
            return None

    def _read_theme(self, section: SectionProxy, key_name: str, fallback: ThemeEnum) -> ThemeEnum:
        try:
            theme = read_str(section, key_name, "").upper()
            if theme == "DARK":
                return ThemeEnum.DARK
            if theme == "LIGHT":
                return ThemeEnum.LIGHT
            if theme == "GENERIC":
                return ThemeEnum.GENERIC

            return fallback

        except Exception:
            log.warning(f"Failed to parse config item: {key_name}, using fallback: {fallback}", exc_info=True)
            return fallback



from dataclasses import dataclass

from PySide6.QtCore import QCoreApplication

from clochette.application.service.provider.dto.ProviderDTO import ProviderDTO
from clochette.provider.public.configuration.PublicURLSourceConfigurationService import (
    PublicURLSourceConfigurationService,
)
from clochette.provider.public.download.PublicCalendarDownloaderService import (
    PublicCalendarDownloaderService,
)
from clochette.provider.public.dto.PublicProviderType import PublicProviderType
from clochette.provider.public.ui.QPublicAuthComponent import QPublicAuthComponent


@dataclass
class PublicProviderFactory:
    """Provider Factory for Public URL calendars."""

    _download_calendar: PublicCalendarDownloaderService
    _provider_configuration_mapper: PublicURLSourceConfigurationService
    _authentication_component: QPublicAuthComponent

    def create(self) -> ProviderDTO:
        """Factory method to create a PublicProviderDTO with all dependencies."""
        return ProviderDTO(
            download_calendar=self._download_calendar,
            provider_type=PublicProviderType(),
            provider_configuration_mapper=self._provider_configuration_mapper,
            display_name=QCoreApplication.translate("PublicProviderFactory", "Public URL"),
            authentication_component=self._authentication_component,
        )



import logging
from dataclasses import dataclass
from logging.handlers import TimedRotatingFileHandler

from clochette import log
from clochette.persist.configuration.ConfigurationLoggingService import ConfigurationLoggingService
from clochette.persist.configuration.PlatformConfigurationService import PlatformConfigurationService


@dataclass
class LoggingService:
    _platform_configuration_service: PlatformConfigurationService
    _configuration_logging_service: ConfigurationLoggingService

    def setup(self) -> None:
        configuration = self._configuration_logging_service.configuration

        # set log level
        log.setLevel(configuration.log_level)
        logging.getLogger("urllib3").setLevel(configuration.log_level)

        # Handler and formatter setup
        formatter = logging.Formatter(
            "%(asctime)s - %(levelname)s - %(filename)s:%(funcName)s():%(lineno)d - %(message)s"
        )

        # File handler with daily rollover (keep only 1 day)
        log_file = self._platform_configuration_service.application_data_path() / "clochette.log"
        file_handler = TimedRotatingFileHandler(str(log_file), when="midnight", interval=1, backupCount=0)
        file_handler.setFormatter(formatter)
        log.addHandler(file_handler)

        # set console handler
        if configuration.log_console_enabled:
            console_handler = logging.StreamHandler()
            console_handler.setFormatter(formatter)
            log.addHandler(console_handler)



from uuid import UUID

from reactivex import Observable
from reactivex.abc import DisposableBase

from clochette import log


class ScheduleStore:
    """Store for managing scheduler subscriptions and tasks."""

    _subscriptions: dict[UUID, DisposableBase]
    _tasks: dict[UUID, Observable[None]]

    def __init__(self):
        self._subscriptions = {}
        self._tasks = {}

    def store_subscription(self, task_id: UUID, subscription: DisposableBase) -> None:
        """Store a subscription with the given task ID."""
        self._subscriptions[task_id] = subscription

    def store_task(self, task_id: UUID, task: Observable[None]) -> None:
        """Store a task Observable with the given task ID."""
        self._tasks[task_id] = task

    def get_subscription(self, task_id: UUID) -> DisposableBase | None:
        """Retrieve a subscription by task ID."""
        return self._subscriptions.get(task_id)

    def get_task(self, task_id: UUID) -> Observable[None] | None:
        """Retrieve a task Observable by task ID."""
        return self._tasks.get(task_id)

    def remove_subscription(self, task_id: UUID) -> DisposableBase | None:
        """Remove and return a subscription by task ID."""
        self._tasks.pop(task_id, None)  # Also remove the task
        return self._subscriptions.pop(task_id, None)

    def get_all_subscriptions(self) -> dict[UUID, DisposableBase]:
        """Get all subscriptions."""
        return self._subscriptions.copy()

    def clear_all_subscriptions(self) -> None:
        """Clear all subscriptions and tasks."""
        for task_id, subscription in self._subscriptions.items():
            try:
                subscription.dispose()
            except Exception:
                log.error(f"Failed to dispose subscription, task_id: {task_id}", exc_info=True)

        self._subscriptions.clear()
        self._tasks.clear()



from PySide6.QtCore import QObject, QEvent
from PySide6.QtGui import Qt, QKeyEvent, QIcon
from PySide6.QtWidgets import QVBoxLayout
from reactivex import Observable
from typing_extensions import override

from clochette.domain.entity.delta.SnoozeDelta import SnoozeDelta
from clochette.framework.qt.Signal import OutSolidSignal, OutHollowSignal, InHollowSignal
from clochette.presentation.widget.QOkCancelButton import QOkCancelButton
from clochette.presentation.widget.QTimedeltaInput import QTimeDeltaInput
from clochette.presentation.window.QAbstractWindow import QAbstractWindow


class QCustomSnoozeWindow(QAbstractWindow):
    on_submitted: OutSolidSignal[SnoozeDelta]
    on_cancelled: OutHollowSignal
    show_window: InHollowSignal

    _input: QTimeDeltaInput

    def __init__(self, icon_window_observable: Observable[QIcon]) -> None:
        super().__init__(icon_window_observable, self.tr("Clochette - Snooze Selection"))

        self.on_submitted = OutSolidSignal()
        self.on_cancelled = OutHollowSignal()
        self.show_window = InHollowSignal(self._show)

        self.setWindowFlag(Qt.WindowType.WindowStaysOnTopHint, True)

        main_layout = QVBoxLayout()
        self.setLayout(main_layout)

        self._input = QTimeDeltaInput()
        main_layout.addWidget(self._input)

        ok_cancel_button = QOkCancelButton()
        ok_cancel_button.on_cancel_clicked.link(self._on_cancel_clicked)
        ok_cancel_button.on_ok_clicked.link(self._on_ok_clicked)
        main_layout.addWidget(ok_cancel_button)

        self.installEventFilter(self)

        self.hide()

    @override
    def eventFilter(self, source: QObject, event: QEvent):
        if isinstance(event, QKeyEvent) and event.key() == Qt.Key.Key_Return:
            self._on_ok_clicked()

        return super().eventFilter(source, event)

    def _on_ok_clicked(self) -> None:
        self.on_submitted.send(SnoozeDelta(self._input.get_time_delta()))
        self._input.clear()
        self.hide()

    def _on_cancel_clicked(self) -> None:
        self.on_cancelled.send()
        self._input.clear()
        self.hide()

    def _show(self):
        super().show()
        self._input.setFocus()



from dataclasses import dataclass
from datetime import timedelta

from clochette.domain.entity.delta.AlarmDelta import AlarmDelta
from clochette.domain.entity.ISourceCalendar import ISourceCalendar
from clochette.domain.entity.configuration.HttpTimeout import HttpTimeout


@dataclass(frozen=True)
class CalendarFormValuesDTO:
    """Simple DTO containing form values from the calendar configuration view"""

    name: str
    source: ISourceCalendar
    download_interval: timedelta
    missed_reminders_past_window: timedelta
    http_timeout: HttpTimeout
    force_alarms: list[AlarmDelta]
    force_alarms_dates: list[AlarmDelta]
    default_alarms: list[AlarmDelta]
    default_alarms_dates: list[AlarmDelta]



from abc import ABC
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from clochette.application.service.provider.dto.ProviderType import ProviderType


class ISourceCalendar(ABC):
    """Base interface for calendar sources

    All implementations must include a 'provider_type: ProviderType' field.
    """

    # Type hint for static analysis - all implementations must have this field
    provider_type: "ProviderType"



from dataclasses import dataclass
from datetime import datetime, UTC

from peewee import AutoField

from clochette.persist.database.field.DateTimeField import DateTimeField
from clochette.persist.database.model.BaseModel import BaseModel


class Calendar(BaseModel):
    id = AutoField()
    last_download = DateTimeField(default=datetime.now(tz=UTC))


@dataclass(frozen=True, eq=True)
class CalendarDTO:
    id: str
    last_download: datetime



from PySide6.QtWidgets import QWidget, QVBoxLayout

from clochette.application.service.provider.dto.IAuthenticationComponent import IAuthenticationComponent
from clochette.application.service.provider.dto.ProviderDTO import ProviderDTO
from clochette.application.service.provider.dto.ProviderType import ProviderType
from clochette.application.store.ProviderStore import ProviderStore
from clochette.domain.entity.CalendarID import CalendarID
from clochette.domain.entity.ISourceCalendar import ISourceCalendar
from clochette.framework.qt.Link import Link
from clochette.framework.qt.Signal import InSolidSignal, InHollowSignal
from clochette.presentation.widget.QDynamicWidgetContainer import QDynamicWidgetContainer


class QAuthenticationSourceView(QWidget):
    """Panel that displays provider-specific authentication UI based on selected provider type"""

    set_selected_provider_type: InSolidSignal[ProviderType]
    clear: InHollowSignal

    _container: QDynamicWidgetContainer[ProviderType]
    _provider_component_map: dict[ProviderType, IAuthenticationComponent]

    def __init__(self, provider_store: ProviderStore):
        super().__init__()

        self.set_selected_provider_type = InSolidSignal(self._set_selected_provider_type)
        self.clear = InHollowSignal(self._clear)

        self._provider_component_map = {}

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self._container = QDynamicWidgetContainer[ProviderType]()
        layout.addWidget(self._container)

        Link(
            observable=provider_store.providers,
            handler=self._on_providers_loaded,
            widget=self,
        )

    def _on_providers_loaded(self, providers_list: list[ProviderDTO]) -> None:
        """Subscribe to provider store to populate authentication panels"""
        # Clear existing mappings
        self._provider_component_map.clear()

        # Add each provider's authentication component
        for provider_dto in providers_list:
            provider_type = provider_dto.provider_type
            self._provider_component_map[provider_type] = provider_dto.authentication_component

            # Add the component's view to the container
            self._container.add_widget(provider_type, provider_dto.authentication_component.view())

    def _set_selected_provider_type(self, provider_type: ProviderType) -> None:
        self._container.show_widget(provider_type)

    def _clear(self):
        """Clear all provider-specific panels"""
        for component in self._provider_component_map.values():
            component.clear()

    def get_selected_provider_type(self) -> ProviderType:
        return self._container.current_key()

    def get_component_by_provider_type(self, provider_type: ProviderType) -> IAuthenticationComponent:
        """Get the authentication component for a specific provider type"""
        return self._provider_component_map[provider_type]

    def get_source(self) -> ISourceCalendar:
        """Get the calendar source from the selected provider panel"""
        provider_type = self.get_selected_provider_type()
        component = self._provider_component_map[provider_type]
        if not component.validate():
            raise ValueError(f"{provider_type.get_id()} authentication validation failed")
        return component.get_source()

    def set_values_from_source(self, calendar_id: CalendarID, source: ISourceCalendar) -> None:
        """Set values in the appropriate provider panel based on source type"""
        for provider_type, component in self._provider_component_map.items():
            if source.provider_type == provider_type:
                component.set_values(calendar_id, source)
                self.set_selected_provider_type(provider_type)
                return
        raise ValueError(f"No provider found for provider type: {source.provider_type}")



from datetime import timedelta, datetime, UTC
from typing import Iterator

from clochette.domain.entity.Trigger import Trigger
from clochette.infrastructure.clock.DateTimeUtils import DateTimeUtils
from clochette.infrastructure.clock.Generics import DateOrDatetimeType
from clochette.infrastructure.data_structure.PeekableIterator import PeekableIterator, take_while


class TriggerIteratorFactory:
    @staticmethod
    def make_trigger_iterator(
        base_iterator: Iterator[Trigger], offsets: list[timedelta], start: datetime
    ) -> PeekableIterator[Trigger]:
        offset_iterator = TriggerIteratorFactory._make_offset_iterator(base_iterator, offsets)
        iterator: PeekableIterator[Trigger] = PeekableIterator(offset_iterator)

        # trim the first values that are before start
        def trim(trigger: Trigger) -> bool:
            return trigger.trigger <= start

        _ = take_while(iterator, trim)
        return iterator

    @staticmethod
    def to_trigger_iterator(base_iterator: Iterator[DateOrDatetimeType]) -> Iterator[Trigger]:
        for trigger in base_iterator:
            yield TriggerIteratorFactory._normalize_trigger(trigger, trigger)

    @staticmethod
    def _make_offset_iterator(base_iterator: Iterator[Trigger], offsets: list[timedelta]) -> Iterator[Trigger]:
        for trigger in base_iterator:
            for offset in offsets:
                yield Trigger(trigger.trigger + offset, trigger.start)

    @staticmethod
    def _normalize_trigger(trigger: DateOrDatetimeType, start: DateOrDatetimeType):
        if DateTimeUtils.is_datetime(trigger):
            if not trigger.tzinfo:
                return Trigger(DateTimeUtils.set_tz(trigger), start)
            else:
                return Trigger(trigger, start)
        else:
            trigger_dt = DateTimeUtils.to_datetime(trigger, UTC)
            return Trigger(trigger_dt, start)



from PySide6.QtGui import QMouseEvent
from PySide6.QtWidgets import QLabel

from clochette.framework.qt.Signal import OutHollowSignal


class QClickableLabel(QLabel):
    on_clicked: OutHollowSignal

    def __init__(self, title: str):
        super().__init__(title)
        self.on_clicked = OutHollowSignal()

    def mousePressEvent(self, event: QMouseEvent) -> None:
        self.on_clicked.send()
        QLabel.mousePressEvent(self, event)



from clochette.provider.shared.dto.AuthCalendar import AuthCalendar


class GoogleCalendar(AuthCalendar):
    _id: str
    _summary: str

    def __init__(self, id: str, summary: str):
        self._id = id
        self._summary = summary

    @property
    def id(self) -> str:
        return self._id

    @property
    def summary(self) -> str:
        return self._summary



from dataclasses import dataclass
from typing import TypeVar, Generic

from PySide6.QtWidgets import QCheckBox, QFormLayout, QGroupBox

_T = TypeVar("_T")


@dataclass(frozen=True, eq=True)
class CheckBoxItem(Generic[_T]):
    display: str
    id: _T


class QCheckBoxForm(QGroupBox, Generic[_T]):
    _layout: QFormLayout
    _checkboxes: dict[CheckBoxItem[_T], QCheckBox]

    def __init__(self, title: str):
        super().__init__(title)

        self._checkboxes = {}

        self._layout = QFormLayout()
        self.setLayout(self._layout)

    def get_checkbox_states(self) -> dict[CheckBoxItem[_T], bool]:
        return {item: checkbox.isChecked() for item, checkbox in self._checkboxes.items()}

    def set_checkbox_states(self, checkbox_states: dict[CheckBoxItem[_T], bool]) -> None:
        self._clear()

        for item, state in checkbox_states.items():
            checkbox = QCheckBox(item.display)
            checkbox.setChecked(state)
            self._layout.addRow(checkbox)
            self._checkboxes[item] = checkbox

        self.adjustSize()

    def _clear(self):
        for _ in self._checkboxes:
            self._layout.removeItem(self._layout.itemAt(0))



from dataclasses import dataclass

from clochette.domain.entity.CalendarID import CalendarID
from clochette.domain.entity.EventUID import EventUID


@dataclass(frozen=True, eq=True)
class EventID:
    event_uid: EventUID
    calendar_id: CalendarID



from PySide6.QtGui import QFont
from PySide6.QtWidgets import QWidget, QLabel, QVBoxLayout

from clochette.presentation.widget.QHLine import QHLine


class QTitle(QWidget):

    _title_label: QLabel

    def __init__(self, title: str, subtitle: str | None = None):
        super().__init__()

        layout = QVBoxLayout()
        self.setLayout(layout)
        layout.setContentsMargins(0, 5, 0, 0)

        self._title_label = QLabel(title)

        title_font = QFont()
        title_font.setPointSize(12)
        self._title_label.setFont(title_font)

        layout.addWidget(self._title_label)

        if subtitle is not None:
            subtitle_label = QLabel(subtitle)
            subtitle_label.setStyleSheet("color: #555555;")
            layout.addWidget(subtitle_label)

        layout.addWidget(QHLine())

    def set_title(self, title: str) -> None:
        """Update the title text dynamically"""
        self._title_label.setText(title)



from dataclasses import dataclass

from clochette.application.service.provider.dto.IAuthenticationComponent import IAuthenticationComponent
from clochette.application.service.provider.dto.IDownloadCalendar import IDownloadCalendar
from clochette.application.service.provider.dto.IProviderConfigurationMapper import IProviderConfigurationMapper
from clochette.application.service.provider.dto.ProviderType import ProviderType


@dataclass
class ProviderDTO:
    download_calendar: IDownloadCalendar
    provider_type: ProviderType
    provider_configuration_mapper: IProviderConfigurationMapper
    display_name: str
    authentication_component: IAuthenticationComponent



from dataclasses import dataclass
from datetime import timedelta, datetime, date, UTC
from typing import Type, Iterator, TypeVar

from icalendar import Calendar, Event, Alarm
from icalendar.cal import Component
from icalendar.prop import vCalAddress, vDDDLists

from clochette import log
from clochette.domain.entity.CalendarEvent import CalendarEvent
from clochette.domain.entity.CalendarEventQuery import CalendarEventQuery
from clochette.domain.entity.CalendarID import CalendarID
from clochette.domain.entity.Contact import Contact
from clochette.domain.entity.EventDetails import EventDetails
from clochette.domain.entity.EventID import EventID
from clochette.domain.entity.EventUID import EventUID
from clochette.domain.entity.Trigger import Trigger
from clochette.domain.entity.configuration.CalendarConfiguration import CalendarConfiguration
from clochette.application.service.ical.RRuleParserService import RRuleParserService
from clochette.application.service.ical.TriggerIteratorFactory import TriggerIteratorFactory
from clochette.application.service.ical.dto.EventDTO import EventDTO
from clochette.application.service.ical.dto.ExclusionDTO import ExclusionDTO
from clochette.infrastructure.clock.DateTimeUtils import DateTimeUtils


@dataclass
class ICalendarService:
    _rrule_parser_service: RRuleParserService

    def parse_calendar(
        self, calendar_configuration: CalendarConfiguration, ics: str, last_download: datetime, calendar_id: CalendarID
    ) -> CalendarEventQuery:
        log.debug("Parsing the ics file")
        cal = Calendar.from_ical(ics)
        events: dict[str, EventDTO] = {}

        # example https://gist.github.com/meskarune/63600e64df56a607efa211b9a87fb443
        for vevent in self._walk(cal, "VEVENT", Event):
            try:
                event = self._get_event(vevent)

                if not event:
                    continue

                if isinstance(event, ExclusionDTO):
                    linked_event = events[event.uid]
                    linked_event.raw += "\n" + event.raw
                    linked_event.exdates.append(event.exclusion)
                    linked_event.rdates.append(event.inclusion)
                else:
                    event.alarms = self._get_alarm(vevent, calendar_configuration)
                    events[event.uid] = event
            except Exception:
                log.warning(f"Failed to parse event {vevent}", exc_info=True)

        res: dict[EventID, CalendarEvent] = {}
        for _, event in events.items():
            calendar_event = self._to_calendar_event(event, calendar_id, last_download)
            res[calendar_event.event_id] = calendar_event

        log.debug(f"Creating a CalendarQueryEvent starting at {last_download}, calendar_id: {calendar_id}")
        return CalendarEventQuery(res)

    def _to_calendar_event(self, event: EventDTO, calendar_id: CalendarID, last_download: datetime) -> CalendarEvent:
        trigger_iterator: Iterator[Trigger]
        log.debug(f"Creating event {event}, calendar_id: {calendar_id}, last_download: {last_download}")

        if event.rrule:
            rrule_iterator = self._rrule_parser_service.parse_rrule(
                event.rrule, event.dtstart, last_download, event.exdates, event.rdates
            )
            trigger_iterator = TriggerIteratorFactory.to_trigger_iterator(rrule_iterator)
        else:
            start: datetime
            if DateTimeUtils.is_datetime(event.dtstart):
                start = DateTimeUtils.set_tz(event.dtstart, UTC)
            else:
                start = DateTimeUtils.to_datetime(event.dtstart, UTC)

            trigger = Trigger(start, event.dtstart)
            trigger_iterator = iter([trigger])

        peekable_trigger_iterator = TriggerIteratorFactory.make_trigger_iterator(
            trigger_iterator, event.alarms, last_download
        )

        return CalendarEvent(
            EventID(EventUID(event.uid), calendar_id), event.dtstart, event.dtend, peekable_trigger_iterator, event.raw
        )

    def _get_event(self, event: Event) -> EventDTO | ExclusionDTO | None:
        rrule = event.get("RRULE")
        uid = str(event.get("UID"))
        dtstart: date | datetime = event.get("DTSTART").dt
        duration = event.get("DURATION")
        recurrence_id = event.get("RECURRENCE-ID")
        rdates = self.get_xdates(event.get("RDATE"))
        exdates = self.get_xdates(event.get("EXDATE"))

        dtend: date | datetime
        if event.get("DTEND"):
            dtend = event.get("DTEND").dt
        elif duration:
            dtend = dtstart + duration.dt
        else:
            dtend = dtstart

        raw = str(event.to_ical().decode())

        if recurrence_id:
            log.debug(f"Found recurrence exception for event with id: {uid}")
            return ExclusionDTO(uid, recurrence_id.dt, dtstart, raw)
        if rrule:
            rrule_str = rrule.to_ical().decode()
            return EventDTO(uid, dtstart, dtend, rrule_str, raw, exdates, rdates)
        else:
            return EventDTO(uid, dtstart, dtend, None, raw, exdates, rdates)

    def get_xdates(self, dates: list[vDDDLists] | None):
        res = []
        if dates:
            for date in dates:
                if date.dts:
                    for dt in date.dts:
                        res.append(dt.dt)
        return res

    def _get_alarm(self, event: Event, calendar_configuration: CalendarConfiguration) -> list[timedelta]:
        res = []

        for valarm in self._walk(event, "VALARM", Alarm):
            alarm = Alarm(valarm)
            res.append(alarm.get("TRIGGER").dt)

        dtstart: date | datetime = event.get("DTSTART").dt
        is_datetime = DateTimeUtils.is_datetime(dtstart)

        if is_datetime:
            force_alarms_td = [ad.get_timedelta() for ad in calendar_configuration.force_alarms]
            unique_alarms = set(res + force_alarms_td)
            if not unique_alarms:
                default_alarms_td = [ad.get_timedelta() for ad in calendar_configuration.default_alarms]
                unique_alarms = set(default_alarms_td)
        else:
            force_alarms_dates_td = [ad.get_timedelta() for ad in calendar_configuration.force_alarms_dates]
            unique_alarms = set(res + force_alarms_dates_td)
            if not unique_alarms:
                default_alarms_dates_td = [ad.get_timedelta() for ad in calendar_configuration.default_alarms_dates]
                unique_alarms = set(default_alarms_dates_td)

        sorted_alarms = list(unique_alarms)
        sorted_alarms.sort()
        return sorted_alarms

    def parse_event_details(self, str_event: str) -> EventDetails:
        event = self._find_event(str_event)

        summary = event.get("SUMMARY")
        description = event.get("DESCRIPTION")
        location = event.get("LOCATION")
        organizer = self._from_vcaladdress(event.get("ORGANIZER")) if event.get("ORGANIZER") else None
        attendee = self._parse_attendees(event.get("ATTENDEE"))
        duration = self._calculate_duration(event)

        return EventDetails(summary, description, location, organizer, attendee, str_event, duration)

    def _calculate_duration(self, event: Event) -> timedelta:
        dtstart_value = event.get("DTSTART")
        dtstart: date | datetime = dtstart_value.dt

        duration_field = event.get("DURATION")
        if duration_field:
            return duration_field.dt

        dtend_value = event.get("DTEND")
        if dtend_value:
            dtend: date | datetime = dtend_value.dt

            # Ensure type consistency and calculate duration with explicit type narrowing
            if DateTimeUtils.is_datetime(dtstart) and DateTimeUtils.is_datetime(dtend):
                return dtend - dtstart
            elif DateTimeUtils.is_date(dtstart) and DateTimeUtils.is_date(dtend):
                return datetime.combine(dtend, datetime.min.time()) - datetime.combine(dtstart, datetime.min.time())

        # No duration or end time specified
        return timedelta(0)

    def _find_event(self, event: str) -> Event:
        cal = Event.from_ical(event)

        for vevent in self._walk(cal, "VEVENT", Event):
            return vevent

        raise ValueError(f"No event found in string: {event}")

    def _from_vcaladdress(self, value: vCalAddress) -> Contact:
        params = value.params
        return Contact(params.get("CN"), str(value).replace("MAILTO:", ""), params.get("PARTSTAT") == "ACCEPTED")

    def _parse_attendees(self, value: list[vCalAddress] | None) -> list[Contact]:
        if value is None:
            return []
        else:
            return [self._from_vcaladdress(x) for x in value]

    _T = TypeVar("_T")

    def _walk(self, cal: Component, component_name: str, component_type: Type[_T]) -> Iterator[_T]:
        for component in cal.walk():
            if component.name == component_name and isinstance(component, component_type):
                yield component



from dataclasses import dataclass
from uuid import uuid4


@dataclass(frozen=True, eq=True)
class CalendarID:
    id: str

    def __post_init__(self):
        if not self.id:
            raise Exception("id cannot be empty")

    @staticmethod
    def new():
        return CalendarID(str(uuid4()))



from clochette.application.utils.DeltaHelper import snooze_display
from clochette.domain.entity.delta.SnoozeDelta import SnoozeDelta
from clochette.framework.qt.Signal import OutSolidSignal, OutHollowSignal
from clochette.presentation.widget.QConditionalScrollComboBox import QConditionalScrollComboBox


class QSnoozeComboBoxView(QConditionalScrollComboBox):
    """Simple view for snooze selection combo box."""

    on_snooze_selected: OutSolidSignal[SnoozeDelta]
    on_custom_snooze_requested: OutHollowSignal

    _is_date: bool

    def __init__(self, snoozes: list[SnoozeDelta], is_date: bool) -> None:
        super().__init__()
        self.on_snooze_selected = OutSolidSignal()
        self.on_custom_snooze_requested = OutHollowSignal()

        self._is_date = is_date

        self.set_snoozes(snoozes)

        super().activated.connect(self._on_activated)

    def set_snoozes(self, snoozes: list[SnoozeDelta]) -> None:
        """Populate the combo box with snooze options."""
        self.clear()
        for snooze in snoozes:
            # having a snooze "x minutes before start" for an all day event doesn't make much sense
            if snooze.is_positive() > 0 or not self._is_date:
                display_text = snooze_display(snooze)
                self.addItem(display_text, snooze)

        # Adding the custom option as the last item
        self.addItem(self.tr("Custom..."))

    def _on_activated(self, index: int) -> None:
        if index == self.count() - 1:
            # Custom snooze selected - emit signal instead of handling directly
            self.on_custom_snooze_requested.send()
        else:
            self.on_snooze_selected.send(self.currentData())



from PySide6.QtWidgets import QWidget, QGroupBox, QLineEdit, QFormLayout, QVBoxLayout

from clochette.application.service.provider.dto.ProviderDTO import ProviderDTO
from clochette.application.service.provider.dto.ProviderType import ProviderType
from clochette.application.store.ProviderStore import ProviderStore
from clochette.framework.qt.Link import Link
from clochette.framework.qt.Signal import InHollowSignal, InSolidSignal
from clochette.presentation.widget.QGenericComboBox import QGenericComboBox
from clochette.presentation.widget.QOkCancelButton import QOkCancelButton
from clochette.presentation.settings.calendar.provider.QAuthenticationSourceComponent import (
    QAuthenticationSourceComponent,
)


class QAuthenticationConfigurationView(QWidget):
    """Panel for configuring calendar authentication (calendar name + authentication source)"""

    set_selected_provider_type: InSolidSignal[ProviderType]
    update_panel: InSolidSignal[ProviderType]
    set_calendar_name: InSolidSignal[str]
    clear: InHollowSignal

    _calendar_name: QLineEdit
    _auth_combo: QGenericComboBox[ProviderType]
    _authentication_source_component: QAuthenticationSourceComponent
    _ok_cancel_button: QOkCancelButton

    def __init__(
        self,
        authentication_source_component: QAuthenticationSourceComponent,
        provider_store: ProviderStore,
    ):
        super().__init__()

        self.set_selected_provider_type = InSolidSignal(
            self._set_selected_provider_type
        )
        self.update_panel = InSolidSignal(self._update_panel)
        self.set_calendar_name = InSolidSignal(self._set_calendar_name)
        self.clear = InHollowSignal(self._clear)

        self._authentication_source_component = authentication_source_component

        layout = QVBoxLayout()
        self.setLayout(layout)

        # Calendar name and authentication type selector
        auth_group = QGroupBox(self.tr("Authentication"))
        auth_layout = QFormLayout()

        self._calendar_name = QLineEdit()
        self._calendar_name.setPlaceholderText(self.tr("Type a name"))
        auth_layout.addRow(self.tr("Calendar Name"), self._calendar_name)

        self._auth_combo = QGenericComboBox[ProviderType]()
        self._auth_combo.currentIndexChanged.connect(self._on_combo_changed)
        auth_layout.addRow(self.tr("Authentication Type"), self._auth_combo)

        Link(
            observable=provider_store.providers,
            handler=self._on_providers_loaded,
            widget=self,
        )

        auth_group.setLayout(auth_layout)
        layout.addWidget(auth_group)

        # Provider-specific authentication panels
        layout.addWidget(self._authentication_source_component.view())

        layout.addStretch()

        self._ok_cancel_button = QOkCancelButton()
        layout.addWidget(self._ok_cancel_button)

    def _on_providers_loaded(self, providers_list: list[ProviderDTO]) -> None:
        """Subscribe to provider store to populate combo box and set default"""
        items = [(p.provider_type, p.display_name) for p in providers_list]
        self._auth_combo.set_items(items)
        # Set default to first provider if items exist
        if items:
            self.update_panel(items[0][0])

    def _on_combo_changed(self) -> None:
        """Handle combo box selection change"""
        current_key = self._auth_combo.current_key()
        if current_key is not None:
            self.update_panel(current_key)

    def _update_panel(self, provider_type: ProviderType) -> None:
        self._authentication_source_component.set_provider_type(provider_type)

    def _set_selected_provider_type(self, provider_type: ProviderType) -> None:
        self._auth_combo.set_current_by_key(provider_type)
        self._authentication_source_component.set_provider_type(provider_type)

    def _set_calendar_name(self, name: str) -> None:
        self._calendar_name.setText(name)

    def _clear(self):
        self._calendar_name.setText("")

    def get_selected_provider_type(self) -> ProviderType:
        return self._authentication_source_component.view().get_selected_provider_type()

    @property
    def calendar_name(self) -> str:
        return self._calendar_name.text()

    @property
    def ok_cancel_button(self) -> QOkCancelButton:
        return self._ok_cancel_button

    @property
    def authentication_source_component(self) -> QAuthenticationSourceComponent:
        return self._authentication_source_component



from reactivex import Observable
from reactivex.subject import ReplaySubject

from clochette.application.service.provider.dto.ProviderDTO import ProviderDTO


class ProviderStore:
    """Store for available provider information."""

    _providers: ReplaySubject[list[ProviderDTO]]

    def __init__(self):
        self._providers = ReplaySubject(buffer_size=1)

    @property
    def providers(self) -> Observable[list[ProviderDTO]]:
        return self._providers

    def set_provider(self, providers: list[ProviderDTO]) -> None:
        self._providers.on_next(providers)



from dataclasses import dataclass

from clochette import log
from clochette.application.usecase.ScheduleCalendarDownloadUseCase import ScheduleCalendarDownloadUseCase
from clochette.domain.entity.configuration.CalendarConfiguration import CalendarConfiguration


@dataclass
class ScheduleAllCalendarsDownloadUseCase:
    """Use case to schedule downloads for all calendars."""

    _schedule_calendar_download_usecase: ScheduleCalendarDownloadUseCase

    def schedule_all_calendars(self, calendars: list[CalendarConfiguration]) -> None:
        """Schedule downloads for all calendars."""
        log.info(f"Scheduling downloads for {len(calendars)} calendars")

        for calendar in calendars:
            self._schedule_calendar_download_usecase.schedule_calendar(calendar)



from PySide6.QtCore import QSize
from PySide6.QtWidgets import QWidget, QVBoxLayout, QPushButton, QStyle, QHBoxLayout

from clochette.framework.qt.Signal import OutHollowSignal


class QCloseablePanel(QWidget):
    on_close_clicked: OutHollowSignal

    def __init__(self, hidden: bool) -> None:
        super().__init__()

        self.on_close_clicked = OutHollowSignal()

        self._layout = QVBoxLayout()

        close_button = QPushButton()
        style = self.style()
        if style:
            close_icon = style.standardIcon(QStyle.StandardPixmap.SP_TitleBarCloseButton)
            close_button.setIcon(close_icon)
            close_button.setIconSize(QSize(16, 16))
            close_button.setFixedSize(QSize(16, 16))

        close_button.clicked.connect(self._on_close_clicked)

        top_panel_layout = QHBoxLayout()
        top_panel_layout.addStretch(1)
        top_panel_layout.addWidget(close_button)

        self._layout.addLayout(top_panel_layout)

        self.setLayout(self._layout)

        if hidden:
            self.hide()

    def _on_close_clicked(self) -> None:
        self.hide()
        self.adjustSize()
        self.on_close_clicked.send()

    def addWidget(self, widget: QWidget) -> None:
        self._layout.addWidget(widget)



from typing import Optional, Callable, TypeVar

from reactivex import Observable, just, throw
from reactivex import operators as ops

from clochette import log

_T = TypeVar("_T")


def retry_with_delay(
    delay_ms: float,
    retry_count: Optional[int] = None,
) -> Callable[[Observable[_T]], Observable[_T]]:
    def delay(ex: Exception, _: Observable[_T]) -> Observable[_T]:
        log.debug("retry_with_delay", exc_info=ex)
        return just(0).pipe(
            ops.delay(delay_ms),
            ops.flat_map(lambda _: throw(ex)),
        )

    def retry(source: Observable[_T]) -> Observable[_T]:
        return source.pipe(
            ops.catch(handler=delay),
            ops.retry(retry_count),
        )

    return retry



from dataclasses import dataclass

from clochette.domain.entity.CalendarID import CalendarID
from clochette.persist.database.dao.CalendarCacheDAO import CalendarCacheDAO


@dataclass
class CalendarCacheRepository:
    _calendar_cache_dao: CalendarCacheDAO

    def get_calendar_cache(self, calendar_id: CalendarID) -> str | None:
        return self._calendar_cache_dao.get_cached_calendar(calendar_id.id)

    def cache_calendar(self, calendar_id: CalendarID, content: str) -> None:
        return self._calendar_cache_dao.cache_calendar(calendar_id.id, content)



from uuid import UUID

from reactivex import Observable, Subject


class SnoozeScheduleStore:
    """Store for managing the snooze reminder schedule task ID using RxPY subjects."""

    _schedule_id: Subject[UUID]

    def __init__(self):
        self._schedule_id = Subject()

    @property
    def schedule_id(self) -> Observable[UUID]:
        """Observable for the schedule ID."""
        return self._schedule_id

    def set_schedule_id(self, task_id: UUID) -> None:
        """Store the scheduler task ID for snooze reminders."""
        self._schedule_id.on_next(task_id)



import configparser
from dataclasses import dataclass
from typing import Any, Sequence, TypeAlias

from isodate import duration_isoformat
from reactivex import Observable, empty, from_iterable, operators as ops

from clochette import log
from clochette.application.service.configurartion.SourceConfigurationMapperStrategy import (
    SourceConfigurationMapperStrategy,
)
from clochette.domain.entity.delta.Delta import Delta
from clochette.domain.entity.configuration.ThemeEnum import ThemeEnum
from clochette.framework.rx.Scheduler import scheduler
from clochette.persist.configuration.PlatformConfigurationService import PlatformConfigurationService
from clochette.persist.configuration.dto.CalendarConfigurationDTO import CalendarConfigurationDTO
from clochette.persist.configuration.dto.GlobalConfigurationDTO import GlobalConfigurationDTO

CalendarParameters: TypeAlias = dict[str, Any]
CalendarConfig: TypeAlias = tuple[str, CalendarParameters]

_CONFIG_SECTION = "Config"


@dataclass
class ConfigurationWriterService:
    _platform_config: PlatformConfigurationService
    _source_configuration_mapper_strategy: SourceConfigurationMapperStrategy

    def write(self, global_configuration: GlobalConfigurationDTO) -> Observable[None]:
        """Write configuration to file, returning an Observable."""
        log.info("Writing Configuration")

        return self._get_all_calendars(global_configuration.calendars).pipe(
            ops.map(lambda calendar_dicts: self._write_config_file(global_configuration, calendar_dicts)),
        )

    def _get_all_calendars(self, calendars: list[CalendarConfigurationDTO]) -> Observable[list[CalendarConfig]]:
        """Get all calendar configurations as an observable list of tuples (id, config)."""
        return from_iterable(calendars, scheduler=scheduler).pipe(
            ops.flat_map(lambda calendar: self._get_calendar_dict(calendar)),
            ops.to_list(),
        )

    def _get_calendar_dict(self, calendar: CalendarConfigurationDTO) -> Observable[CalendarConfig]:
        """Get calendar configuration dict with error handling."""

        return self._source_configuration_mapper_strategy.write(calendar.source).pipe(
            ops.map(lambda x: self._build_calendar_dict(calendar, x)),
            ops.catch(lambda error, _src: self._handle_write_error(calendar, error)),
        )

    def _build_calendar_dict(
        self, calendar: CalendarConfigurationDTO, authentication: dict[str, str]
    ) -> CalendarConfig:
        """Build calendar configuration dictionary from authentication and calendar data."""
        calendar_dict = authentication | {
            "name": calendar.name,
            "download.interval.minutes": int(calendar.download_interval.total_seconds() / 60),
            "missed.reminders.past_window": duration_isoformat(calendar.missed_reminders_past_window),
            "force_alarms": self._get_timedelta_list(calendar.force_alarms),
            "force_alarms_dates": self._get_timedelta_list(calendar.force_alarms_dates),
            "default_alarms": self._get_timedelta_list(calendar.default_alarms),
            "default_alarms_dates": self._get_timedelta_list(calendar.default_alarms_dates),
            "timeout.connection": calendar.http_timeout.connection_timeout,
            "timeout.read": calendar.http_timeout.read_timeout,
        }
        return (calendar.id.id, calendar_dict)

    def _handle_write_error(
        self, calendar: CalendarConfigurationDTO, error: Exception
    ) -> Observable[tuple[str, dict[str, str]]]:
        """Handle errors from writing calendar configuration by logging and returning empty."""
        log.warning(f"Failed to write calendar configuration: {calendar.id.id}", exc_info=error)
        return empty()

    def _write_config_file(
        self, global_configuration: GlobalConfigurationDTO, calendar_dicts: list[CalendarConfig]
    ) -> None:
        """Write the complete configuration file with global settings and all calendars."""
        config = configparser.RawConfigParser()

        snoozes_str = self._get_timedelta_list(global_configuration.snoozes)
        theme_icon_window = self._get_theme(global_configuration.theme.window_icon_theme)
        systray_icon_theme = self._get_theme(global_configuration.theme.systray_icon_theme)

        config[_CONFIG_SECTION] = {
            "snooze": snoozes_str,
            "theme.icon.window": theme_icon_window,
            "theme.icon.systray": systray_icon_theme,
        }

        for calendar_id, calendar_config in calendar_dicts:
            config[calendar_id] = calendar_config

        configuration_file = self._platform_config.application_configuration_file()

        with open(configuration_file, "w") as configfile:
            config.write(configfile)

    def _get_theme(self, theme: ThemeEnum) -> str:
        if theme == ThemeEnum.DARK:
            return "DARK"
        if theme == ThemeEnum.LIGHT:
            return "LIGHT"
        return "GENERIC"

    def _get_timedelta(self, value: Delta) -> str:
        return duration_isoformat(value.get_timedelta())

    def _get_timedelta_list(self, values: Sequence[Delta]) -> str:
        return ",".join(list(map(self._get_timedelta, values)))



import json
from dataclasses import dataclass
from http.server import BaseHTTPRequestHandler

from oauthlib.oauth2 import OAuth2Token
from reactivex import Subject, Observable, defer
from reactivex import operators as ops

from clochette import log
from clochette.infrastructure.clock.ClockService import ClockService
from clochette.infrastructure.http_.client.HttpService import HttpService
from clochette.infrastructure.http_.server.RequestHandlerFactory import RequestHandlerFactory
from clochette.infrastructure.http_.server.SingleRequestHttpsServer import SingleRequestHttpsServer
from clochette.presentation.window.QWebBrowserWindow import QWebBrowserWindow


@dataclass(frozen=True)
class OAuth2Service:
    _http_service: HttpService
    _clock_service: ClockService
    _web_browser_window: QWebBrowserWindow
    _port: int = 0

    def refresh_token(self, remote_server_url: str, token: OAuth2Token) -> OAuth2Token:
        headers = {"Content-Type": "application/json"}

        body = {"action": "refresh_token", "refresh_token": token.get("refresh_token")}

        response = self._http_service.post(remote_server_url, json.dumps(body), headers)

        json_response = response.content_json
        if json_response is None or not response.is_successful():
            raise Exception(f"Failed to refresh token, http status: {response.status_code}")
        else:
            return OAuth2Token(json_response)

    def authenticate(self, remote_server_url: str) -> Observable[OAuth2Token]:
        return defer(lambda _: self._authenticate(remote_server_url))

    def _authenticate(self, remote_server_url: str):
        log.info("Retrieving a valid token")
        authorization_subject: Subject[str] = Subject()

        def do_get(request_handler: BaseHTTPRequestHandler) -> None:
            log.debug("Authorization received")
            request_handler.send_response(200)
            request_handler.send_header("Content-type", "text/html")
            request_handler.end_headers()
            request_handler.wfile.write(b"Oauth2 callback received")
            authorization_subject.on_next(request_handler.path)

        http_server = SingleRequestHttpsServer(RequestHandlerFactory(do_get), self._port)
        http_server.start()

        redirect_uri = http_server.get_url() + "/callback"

        location = self._retrieve_location(remote_server_url, redirect_uri)
        log.debug("Location retrieved, opening Web Browser")
        self._web_browser_window.show_browser()
        self._web_browser_window.load_url(location)
        self._web_browser_window.on_browser_closed.link(
            lambda: authorization_subject.on_error(Exception("Closed the browser before authenticating"))
        )

        def cleanup() -> None:
            http_server.stop()
            self._web_browser_window.hide_browser()

        return authorization_subject.pipe(
            ops.do_action(lambda _: cleanup()),
            ops.map(lambda r: self._fetch_token(remote_server_url, r, redirect_uri)),
        )

    def _retrieve_location(self, remote_server_url: str, redirect_uri: str) -> str:
        log.info(f"Authenticating to URI: {remote_server_url}")
        headers = {"Content-Type": "application/json"}

        body = {"action": "authenticate", "redirect_uri": redirect_uri}

        response = self._http_service.post(remote_server_url, json.dumps(body), headers)
        location = response.headers.get("location", None)

        if not location:
            raise Exception("No location received during authentication")

        return location

    def _fetch_token(self, remote_server_url: str, authorization_response: str, redirect_uri: str) -> OAuth2Token:
        log.info("Fetching token")
        headers = {"Content-Type": "application/json"}

        body = {"action": "fetch_token", "authorization_response": authorization_response, "redirect_uri": redirect_uri}
        response = self._http_service.post(remote_server_url, json.dumps(body), headers)

        json_response = response.content_json
        if json_response is None or not response.is_successful():
            raise Exception(f"Failed to authenticate, http status: {response.status_code}")
        else:
            return OAuth2Token(json_response)

    def has_expired(self, token: OAuth2Token) -> bool:
        return self._clock_service.epoch() > int(token["expires_at"])



from __future__ import annotations

from dataclasses import dataclass
from datetime import timedelta

from clochette.domain.entity.delta.AlarmDelta import AlarmDelta
from clochette.domain.entity.CalendarID import CalendarID
from clochette.domain.entity.configuration.HttpTimeout import HttpTimeout
from clochette.domain.entity.ISourceCalendar import ISourceCalendar


@dataclass(frozen=True, eq=True)
class CalendarConfigurationDTO:
    id: CalendarID
    name: str
    source: ISourceCalendar
    force_alarms: list[AlarmDelta]
    force_alarms_dates: list[AlarmDelta]
    default_alarms: list[AlarmDelta]
    default_alarms_dates: list[AlarmDelta]
    download_interval: timedelta
    missed_reminders_past_window: timedelta
    http_timeout: HttpTimeout

    def __post_init__(self):
        if self.download_interval.total_seconds() <= 0:
            raise ValueError(f"download_interval cannot be null or negative, calendar: {self.name}")

        if self.missed_reminders_past_window.total_seconds() > 0:
            raise ValueError(f"missed_reminders_past_window cannot be positive, calendar: {self.name}")



from datetime import timedelta

from peewee import Field


class TimedeltaField(Field):
    field_type = "float"

    def db_value(self, value: timedelta) -> float:
        return value.total_seconds()

    def python_value(self, value: float) -> timedelta:
        return timedelta(seconds=value)



from typing import Generic, TypeVar

from PySide6.QtWidgets import QComboBox

_T = TypeVar("_T")


class QTypedComboBox(QComboBox, Generic[_T]):
    """
    A generic combobox that stores typed values alongside display text.
    Each item has a display value (shown to user) and a typed value (stored data).
    """

    def __init__(self):
        super().__init__()

    def add_item(self, display_value: str, value: _T) -> None:
        """Add an item with display text and associated typed value"""
        self.addItem(display_value, value)

    def add_items(self, items: list[tuple[str, _T]]) -> None:
        """Add multiple items at once. Each item is a tuple of (display_value, value)"""
        for display_value, value in items:
            self.add_item(display_value, value)

    def get_current_value(self) -> _T | None:
        """Get the typed value of the currently selected item"""
        return self.currentData()

    def set_current_by_value(self, value: _T) -> bool:
        """
        Set the current item by its typed value.
        Returns True if found and set, False otherwise.
        """
        for i in range(self.count()):
            if self.itemData(i) == value:
                self.setCurrentIndex(i)
                return True
        return False

    def set_current_by_display(self, display_value: str) -> bool:
        """
        Set the current item by its display text.
        Returns True if found and set, False otherwise.
        """
        for i in range(self.count()):
            if self.itemText(i) == display_value:
                self.setCurrentIndex(i)
                return True
        return False



from dataclasses import dataclass

from PySide6.QtCore import QCoreApplication

from clochette.application.service.provider.dto.ProviderDTO import ProviderDTO
from clochette.provider.basic.configuration.BasicAuthURLSourceConfigurationService import (
    BasicAuthURLSourceConfigurationService,
)
from clochette.provider.basic.download.BasicAuthCalendarDownloaderService import (
    BasicAuthCalendarDownloaderService,
)
from clochette.provider.basic.dto.BasicProviderType import BasicProviderType
from clochette.provider.basic.ui.QBasicAuthComponent import QBasicAuthComponent


@dataclass
class BasicProviderFactory:
    """Provider Factory for Basic Authentication calendars."""

    _download_calendar: BasicAuthCalendarDownloaderService
    _provider_configuration_mapper: BasicAuthURLSourceConfigurationService
    _authentication_component: QBasicAuthComponent

    def create(self) -> ProviderDTO:
        """Factory method to create a BasicProviderDTO with all dependencies."""
        return ProviderDTO(
            download_calendar=self._download_calendar,
            provider_type=BasicProviderType(),
            provider_configuration_mapper=self._provider_configuration_mapper,
            display_name=QCoreApplication.translate("BasicProviderFactory", "Basic Auth"),
            authentication_component=self._authentication_component,
        )



from PySide6.QtWidgets import QFormLayout, QLineEdit, QCheckBox, QPushButton, QLabel, QHBoxLayout, QGroupBox

from clochette.framework.qt.Signal import OutHollowSignal, InHollowSignal, InSolidSignal
from clochette.presentation.widget.QWidgetLayout import QWidgetLayout


class QBasicAuthView(QGroupBox):

    on_authenticate: OutHollowSignal

    set_authentication_failed: InSolidSignal[Exception]
    set_authentication_success: InHollowSignal
    clear: InHollowSignal
    set_url: InSolidSignal[str]
    set_username: InSolidSignal[str]
    set_password: InSolidSignal[str]
    set_error: InSolidSignal[str]

    _url: QLineEdit
    _username: QLineEdit
    _password: QLineEdit
    _auth_button: QPushButton
    _error_label: QLabel

    def __init__(self):
        super().__init__(self.tr("Basic Authentication Options"))

        self.on_authenticate = OutHollowSignal()

        self.set_authentication_failed = InSolidSignal(self._authentication_failed)
        self.set_authentication_success = InHollowSignal(self._authentication_success)
        self.clear = InHollowSignal(self._clear)
        self.set_url = InSolidSignal(self._set_url)
        self.set_username = InSolidSignal(self._set_username)
        self.set_password = InSolidSignal(self._set_password)
        self.set_error = InSolidSignal(self._set_error)

        layout = QFormLayout()
        self.setLayout(layout)

        self._url = QLineEdit()
        self._url.setPlaceholderText(self.tr("Enter URL"))
        layout.addRow(self.tr("URL:"), self._url)

        self._username = QLineEdit()
        self._username.setPlaceholderText(self.tr("Enter username"))
        layout.addRow(self.tr("Username:"), self._username)

        self._password = QLineEdit()
        self._password.setPlaceholderText(self.tr("Enter password"))
        layout.addRow(self.tr("Password:"), self._password)
        self._password.setEchoMode(QLineEdit.EchoMode.Password)

        show_password_checkbox = QCheckBox(self.tr("Show Password"))
        show_password_checkbox.stateChanged.connect(
            lambda state: (
                self._password.setEchoMode(QLineEdit.EchoMode.Normal)
                if state
                else self._password.setEchoMode(QLineEdit.EchoMode.Password)
            )
        )

        layout.addRow("", show_password_checkbox)

        button_layout = QWidgetLayout(QHBoxLayout())
        button_layout.add_stretch()

        self._auth_button = QPushButton(self.tr("Authenticate"))
        self._auth_button.clicked.connect(self._authenticate)
        button_layout.add_widget(self._auth_button)

        layout.addRow(button_layout)

        self._error_label = QLabel()
        self._error_label.setWordWrap(True)
        layout.addRow(self._error_label)

    def _authenticate(self):
        self._auth_button.setText(self.tr("Authenticating..."))
        self._auth_button.setEnabled(False)
        self.on_authenticate.send()

    def _authentication_failed(self, e: Exception):
        self._error_label.setStyleSheet("color: red;")
        self._error_label.setText(self.tr("Failed to authenticate: %s") % e)
        self._auth_button.setText(self.tr("Authenticate"))
        self._auth_button.setEnabled(True)

    def _authentication_success(self):
        self._error_label.setStyleSheet("color: green;")
        self._error_label.setText(self.tr("Authentication successful"))
        self._auth_button.setText(self.tr("Authenticate"))
        self._auth_button.setEnabled(True)

    def _set_url(self, url: str) -> None:
        self._url.setText(url)

    def _set_username(self, username: str) -> None:
        self._username.setText(username)

    def _set_password(self, password: str) -> None:
        self._password.setText(password)

    def _set_error(self, error: str) -> None:
        self._error_label.setText(error)

    def _clear(self):
        self._url.setText("")
        self._username.setText("")
        self._password.setText("")
        self._error_label.setText("")

    @property
    def url(self) -> str:
        return self._url.text()

    @property
    def username(self) -> str:
        return self._username.text()

    @property
    def password(self) -> str:
        return self._password.text()

    @property
    def error(self) -> str:
        return self._error_label.text()



from datetime import datetime, date
from typing import Iterator

from dateutil.rrule import rrulestr, rruleset

from clochette import log
from clochette.infrastructure.clock.DateTimeUtils import DateTimeUtils


class RRuleParserService:

    # doc: https://dateutil.readthedocs.io/en/stable/rrule.html
    def parse_rrule(
        self,
        rule: str,
        dtstart: date | datetime,
        after: datetime,
        exclusion: list[datetime] | None = None,
        inclusion: list[datetime] | None = None,
    ) -> Iterator[date] | Iterator[datetime]:
        log.debug(f"Parsing rrule: {rule}, dtstart: {dtstart}, after: {after}")

        exclusion = exclusion or []
        inclusion = inclusion or []

        if DateTimeUtils.is_datetime(dtstart):
            return self._parse_rrule_datetime(rule, dtstart, after, exclusion, inclusion)
        else:
            return self._parse_rrule_date(rule, dtstart, after, exclusion, inclusion)

    def _parse_rrule_date(
        self,
        rule: str,
        dtstart: date,
        after: datetime,
        exclusion: list[datetime],
        inclusion: list[datetime],
    ) -> Iterator[date]:
        after = DateTimeUtils.set_tz(after, None)
        after = DateTimeUtils.to_datetime(after.date())

        ruleset = rruleset()
        ruleset.rrule(rrulestr(rule, dtstart=dtstart))  # pyright: ignore [reportArgumentType]

        for ex in exclusion:
            ruleset.exdate(DateTimeUtils.to_datetime(ex))

        for inc in inclusion:
            ruleset.rdate(DateTimeUtils.to_datetime(inc))

        for res in ruleset.xafter(after, inc=True):
            yield res.date()

    def _parse_rrule_datetime(
        self,
        rule: str,
        dtstart: datetime,
        after: datetime,
        exclusion: list[datetime],
        inclusion: list[datetime],
    ) -> Iterator[datetime]:

        original_tz = dtstart.tzinfo
        until: datetime | None = rrulestr(rule)._until  # pyright: ignore [reportAttributeAccessIssue]

        if until:
            dtstart, after = DateTimeUtils.normalize_tz(until.tzinfo, [dtstart, after])
        else:
            after = DateTimeUtils.set_tz(after, dtstart.tzinfo)

        ruleset = rruleset()
        ruleset.rrule(rrulestr(rule, dtstart=dtstart))  # pyright: ignore [reportArgumentType]

        for ex in exclusion:
            ex = DateTimeUtils.set_tz(ex, dtstart.tzinfo)
            ruleset.exdate(ex)

        for inc in inclusion:
            inc = DateTimeUtils.set_tz(inc, dtstart.tzinfo)
            ruleset.rdate(inc)

        for event in ruleset.xafter(after, inc=True):
            if until:
                yield DateTimeUtils.set_tz(event, original_tz)
            else:
                yield event



from PySide6.QtCore import Qt
from PySide6.QtWidgets import QScrollArea, QWidget, QStyle
from typing_extensions import override


class QVScrollArea(QScrollArea):
    """Vertical-only scroll area with automatic width management to prevent scrollbar overlap"""

    def __init__(self):
        super().__init__()
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setFrameShape(QScrollArea.Shape.NoFrame)

    @override
    def setWidget(self, widget: QWidget) -> None:
        """Set the widget and automatically adjust minimum width to account for scrollbar"""
        super().setWidget(widget)

        # Calculate minimum width: widget's size hint + scrollbar width
        scrollbar_width = self.style().pixelMetric(QStyle.PixelMetric.PM_ScrollBarExtent)
        self.setMinimumWidth(widget.sizeHint().width() + scrollbar_width)



import configparser
from dataclasses import dataclass
from functools import cached_property

from clochette import log
from clochette.domain.entity.configuration.LoggingConfiguration import LoggingConfiguration
from clochette.persist.configuration.ConfigurationMigrationService import ConfigurationMigrationService
from clochette.persist.configuration.ConfigurationReadUtils import read_str, read_bool
from clochette.persist.configuration.PlatformConfigurationService import PlatformConfigurationService


@dataclass
class ConfigurationLoggingService:
    _platform_config: PlatformConfigurationService
    _configuration_migration_service: ConfigurationMigrationService
    CONFIG_SECTION = "Config"

    @cached_property
    def configuration(self) -> LoggingConfiguration:
        configuration_file = self._platform_config.logging_configuration_file()

        try:
            self._configuration_migration_service.init_logging_configuration()
            log.info(f"Start parsing configuration file: {configuration_file}", exc_info=True)
            configuration = configparser.ConfigParser(inline_comment_prefixes=("#", ";"), interpolation=None)
            configuration.read(configuration_file)
            section = configuration[ConfigurationLoggingService.CONFIG_SECTION]

            # logs config
            log_level = read_str(section, "log.level", "INFO").upper()
            enable_console_log = read_bool(section, "log.console.enabled", False)

            return LoggingConfiguration(log_level, enable_console_log)

        except Exception as e:
            log.error(f"Failed to parse config file: {configuration_file}", exc_info=True)
            raise e



import signal
import sys

from PySide6.QtCore import QTranslator
from PySide6.QtWidgets import QApplication

from clochette import log
from clochette.application.i18n.I18nService import I18nService
from clochette.res import resource_rc


class SysArgvProvider:
    def __call__(self) -> list[str]:
        return sys.argv


class QMainApplication:
    _application: QApplication
    _translator: QTranslator
    _i18n_service: I18nService

    def __init__(self, sys_argv_provider: SysArgvProvider, i18n_service: I18nService) -> None:
        _ = resource_rc.qt_resource_data
        # https://stackoverflow.com/questions/5160577/ctrl-c-doesnt-work-with-pyqt
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        self._application = QApplication(sys_argv_provider())

        self._translator = QTranslator()
        self._application.installTranslator(self._translator)

        i18n_service.load_translations(self._translator)

    def stop(self) -> None:
        log.info("Stopping QT application")
        self._application.exit(1)

    def exec(self) -> None:
        """Start the application"""
        log.info("Starting QT application")
        self._application.exec()



from datetime import timedelta
from typing import TypeAlias, Literal

from PySide6.QtCore import QCoreApplication

from clochette.application.i18n.LangHelper import precise_delta
from clochette.domain.entity.delta.AlarmDelta import AlarmDelta
from clochette.domain.entity.delta.Delta import Delta
from clochette.domain.entity.delta.SnoozeDelta import SnoozeDelta

TimeUnitLiteral: TypeAlias = Literal["minutes", "hours", "days"]

time_unit_enum: tuple[TimeUnitLiteral, ...] = ("minutes", "hours", "days")


def make_timedelta(time: int, unit: TimeUnitLiteral) -> timedelta:
    """
    Create a timedelta object from a numeric value and a time unit.

    Parameters:
    - time: int, the amount of time.
    - unit: str, the unit of time ("minutes", "hours", "days").

    Returns:
    - timedelta object representing the duration.
    """
    if unit == "minutes":
        return timedelta(minutes=time)
    elif unit == "hours":
        return timedelta(hours=time)
    else:  # unit == "days"
        return timedelta(days=time)


def snooze_display(delta: SnoozeDelta) -> str:
    """Compute display string for snooze delta."""
    timedelta_value = delta.get_timedelta()
    if timedelta_value.total_seconds() == 0:
        return QCoreApplication.translate("DeltaHelper", "at start")

    precise = precise_delta(timedelta_value)
    if timedelta_value.total_seconds() < 0:
        return QCoreApplication.translate("DeltaHelper", "%s before start") % precise
    else:
        return QCoreApplication.translate("DeltaHelper", "for %s") % precise


def alarm_display(delta: AlarmDelta) -> str:
    """Compute display string for alarm delta."""
    timedelta_value = delta.get_timedelta()
    if timedelta_value.total_seconds() == 0:
        return QCoreApplication.translate("DeltaHelper", "at start")

    precise = precise_delta(timedelta_value)
    if timedelta_value.total_seconds() < 0:
        return QCoreApplication.translate("DeltaHelper", "%s before start") % precise
    else:
        return QCoreApplication.translate("DeltaHelper", "%s after start") % precise


def delta_display(delta: Delta) -> str:
    """Get display text for a delta using DeltaHelper."""
    if isinstance(delta, SnoozeDelta):
        return snooze_display(delta)
    elif isinstance(delta, AlarmDelta):
        return alarm_display(delta)
    else:
        # Fallback for unknown delta types
        return str(delta.get_timedelta())



from dataclasses import dataclass

from clochette import log
from clochette.application.service.SnoozeModel import SnoozeModel
from clochette.application.store.OccurrenceStore import OccurrenceStore
from clochette.domain.entity.Occurrence import Occurrence
from clochette.domain.entity.Snooze import Snooze
from clochette.domain.entity.delta.SnoozeDelta import SnoozeDelta
from clochette.domain.entity.Trigger import Trigger
from clochette.application.repository.SnoozeRepository import SnoozeRepository


@dataclass
class SnoozeOccurrenceUseCase:
    """Use case for snoozing an occurrence."""

    _occurrence_store: OccurrenceStore
    _snooze_model: SnoozeModel
    _snooze_repository: SnoozeRepository

    def snooze(self, occurrence: Occurrence, delta: SnoozeDelta) -> None:
        """Snooze an occurrence by dismissing it and scheduling it for later.

        Args:
            occurrence: The occurrence to snooze
            delta: How long to snooze for
        """
        log.info(f"Snoozed occurrence: {occurrence.event_id}")

        # Dismiss the current occurrence
        self._occurrence_store.dismiss_occurrence(occurrence)

        # Delete any existing snooze for this event
        self._snooze_repository.delete_snoozes([occurrence.event_id])

        # Create new snooze with updated trigger time
        trigger = Trigger(occurrence.trigger.trigger + delta.get_timedelta(), occurrence.trigger.start)
        snooze = Snooze(-1, occurrence.event_id, trigger)

        # Add to both model (in-memory) and repository (database)
        self._snooze_model.add_snooze(snooze)
        self._snooze_repository.add_snooze(snooze)



from datetime import datetime

from clochette.persist.database.model.Calendar import Calendar, CalendarDTO


class CalendarDAO:

    def delete_by_id(self, id: str) -> None:
        query = Calendar.delete().where(Calendar.id == id)
        query.execute()

    def insert(self, id: str) -> None:
        Calendar.insert(id=id).execute()

    def get_all(self) -> list[CalendarDTO]:
        calendars = Calendar.select()
        return [CalendarDTO(id=calendar.id, last_download=calendar.last_download) for calendar in calendars]

    def update_last_download(self, id: str, last_download: datetime) -> None:
        query = Calendar.update(last_download=last_download).where(Calendar.id == id)
        query.execute()

    def get_by_id(self, id: str) -> CalendarDTO:
        calendar = Calendar.get(Calendar.id == id)
        return CalendarDTO(id=calendar.id, last_download=calendar.last_download)



from PySide6.QtWidgets import QFrame


class QVLine(QFrame):
    def __init__(self):

        super().__init__()
        self.setFrameShape(QFrame.Shape.VLine)
        self.setFrameShadow(QFrame.Shadow.Sunken)



from PySide6.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout

from clochette.domain.entity.delta.SnoozeDelta import SnoozeDelta
from clochette.domain.entity.configuration.ThemeConfiguration import ThemeConfiguration
from clochette.domain.entity.configuration.ThemeEnum import ThemeEnum
from clochette.framework.qt.Signal import OutSolidSignal
from clochette.presentation.widget.QThemeRadioButton import QThemeRadioButton
from clochette.presentation.widget.QTimedeltaListWidget import QTimedeltaListWidget
from clochette.presentation.widget.QTitle import QTitle


class QGeneralSettingsView(QWidget):
    """Pure view for general settings - snoozes and theme configuration."""

    # Out signals
    on_snoozes_changed: OutSolidSignal[list[SnoozeDelta]]
    on_theme_changed: OutSolidSignal[ThemeConfiguration]

    _snooze_widget: QTimedeltaListWidget[SnoozeDelta]
    _window_icon_theme_radio_button: QThemeRadioButton
    _systray_icon_theme_radio_button: QThemeRadioButton

    def __init__(self):
        super().__init__()

        # Initialize signals
        self.on_snoozes_changed = OutSolidSignal()
        self.on_theme_changed = OutSolidSignal()

        main_layout = QVBoxLayout()
        self.setLayout(main_layout)

        # Snoozes section
        main_layout.addWidget(
            QTitle(self.tr("Snoozes"), self.tr("Configure the different snooze option when a reminder appears"))
        )

        self._snooze_widget = QTimedeltaListWidget(SnoozeDelta)
        main_layout.addWidget(self._snooze_widget)

        # Wire snooze widget to signal
        self._snooze_widget.on_delta_updated.link(self._on_snoozes_updated)

        # Theme section
        main_layout.addWidget(
            QTitle(self.tr("Theme"), self.tr("Clochette cannot detect the theme so I let you choose it"))
        )

        theme_layout = QHBoxLayout()

        self._window_icon_theme_radio_button = QThemeRadioButton(self.tr("Window Icon Theme"))
        self._systray_icon_theme_radio_button = QThemeRadioButton(self.tr("System Tray Icon Theme"))

        theme_layout.addWidget(self._window_icon_theme_radio_button)
        theme_layout.addWidget(self._systray_icon_theme_radio_button)
        theme_layout.addStretch()

        main_layout.addLayout(theme_layout)

        # Wire theme widgets to signal
        self._window_icon_theme_radio_button.on_theme_changed.link(self._on_any_theme_changed)
        self._systray_icon_theme_radio_button.on_theme_changed.link(self._on_any_theme_changed)

    def set_snoozes(self, snooze_displays: list[SnoozeDelta]) -> None:
        """Set the snooze deltas."""
        self._snooze_widget.set_deltas(snooze_displays)

    def set_window_theme(self, theme: ThemeEnum) -> None:
        """Set the window icon theme."""
        self._window_icon_theme_radio_button.set_theme(theme)

    def set_systray_theme(self, theme: ThemeEnum) -> None:
        """Set the systray icon theme."""
        self._systray_icon_theme_radio_button.set_theme(theme)

    def _on_snoozes_updated(self, snoozes: list[SnoozeDelta]) -> None:
        """Handle snooze updates from widget."""
        self.on_snoozes_changed.send(snoozes)

    def _on_any_theme_changed(self, _) -> None:
        """Handle theme change from either radio button - emit both themes."""
        window_theme = self._window_icon_theme_radio_button.get_theme()
        systray_theme = self._systray_icon_theme_radio_button.get_theme()

        if window_theme is not None and systray_theme is not None:
            self.on_theme_changed.send(ThemeConfiguration(window_theme, systray_theme))



from dataclasses import dataclass
from urllib.parse import quote

from oauthlib.oauth2 import OAuth2Token
from reactivex import Observable, operators as op, just

from clochette import log
from clochette.domain.entity.CalendarID import CalendarID
from clochette.domain.entity.configuration.HttpTimeout import HttpTimeout
from clochette.infrastructure.http_.client.HttpService import HttpService
from clochette.provider.google.download.GoogleOAuth2Service import GoogleOAuth2Service
from clochette.provider.google.dto.GoogleCalendar import GoogleCalendar
from clochette.provider.google.dto.GoogleOauth2DTO import GoogleOauth2DTO
from clochette.provider.google.keyring_.GoogleAuthenticationService import GoogleAuthenticationService
from clochette.provider.shared.oauth2.OAuth2Mapper import OAuth2Mapper


@dataclass
class GoogleService:
    _google_oauth2_service: GoogleOAuth2Service
    _oauth2_mapper: OAuth2Mapper
    _http_service: HttpService
    _google_authentication_service: GoogleAuthenticationService

    def _authenticate(self, calendar_id: CalendarID) -> Observable[OAuth2Token]:
        def store_token(token: OAuth2Token) -> None:
            log.debug("Storing Google token")
            token_str = self._oauth2_mapper.from_oauth2_token(token)

            self._google_authentication_service.store_auth(calendar_id, GoogleOauth2DTO(token_str))

        def retrieve(auth: GoogleOauth2DTO | None) -> Observable[OAuth2Token]:
            if auth is None:
                return self._google_oauth2_service.authenticate().pipe(
                    op.do_action(lambda t: store_token(t)),
                )
            else:
                return just(self._oauth2_mapper.to_oauth2_token(auth.token))

        def refresh_if_needed(token: OAuth2Token) -> OAuth2Token:
            if self._google_oauth2_service.has_expired(token):
                log.info("Google token has expired, refreshing")
                token = self._google_oauth2_service.refresh_token(token)
                store_token(token)
            return token

        return just(self._google_authentication_service.retrieve_auth(calendar_id)).pipe(
            op.flat_map(retrieve),
            op.map(refresh_if_needed),
        )

    def list_calendars(self, calendar_id: CalendarID, http_timeout: HttpTimeout) -> Observable[list[GoogleCalendar]]:
        def download(token: OAuth2Token) -> list[GoogleCalendar]:
            headers = {"Authorization": f"Bearer {token.get("access_token")}", "Accept": "application/json"}

            response = self._http_service.get(
                "https://www.googleapis.com/calendar/v3/users/me/calendarList", headers, http_timeout
            )

            json = response.content_json

            if json is None or not response.is_successful():
                raise Exception(f"failed to list calendars, http status: {response.status_code}")
            else:
                items = json.get("items", [])
                return [GoogleCalendar(item.get("id"), item.get("summary")) for item in items]

        return self._authenticate(calendar_id).pipe(
            op.map(download),
        )

    def download_calendar(
        self, calendar_id: CalendarID, cal: GoogleCalendar, http_timeout: HttpTimeout
    ) -> Observable[str]:
        def download(token: OAuth2Token) -> str:
            headers = {"Authorization": f"Bearer {token.get("access_token")}"}

            id = quote(cal.id)
            response = self._http_service.get(
                f"https://apidata.googleusercontent.com/caldav/v2/{id}/events", headers, http_timeout
            )

            if response.is_successful():
                return response.content_utf8 or ""
            else:
                raise Exception(f"Failed to download calendar: {cal.id}, https status: {response.status_code}")

        return self._authenticate(calendar_id).pipe(op.map(download))



from clochette.provider.microsoft.dto.MicrosoftCalendar import MicrosoftCalendar
from clochette.provider.shared.ui.QOAuth2Panel import QOAuth2Panel


class QMicrosoftOAuth2Panel(QOAuth2Panel[MicrosoftCalendar]):
    pass



from typing import Callable, TypeVar, Generic

from PySide6.QtCore import Signal, QObject

_T = TypeVar("_T")


class _SolidSignal(QObject):
    _signal = Signal(object)


class _HollowSignal(QObject):
    _signal = Signal()


class InHollowSignal:
    _signal: _HollowSignal

    def __init__(self, slot: Callable[[], None] | None = None):
        self._signal = _HollowSignal()
        if slot is not None:
            self.link(slot)

    def __call__(self) -> None:
        self._signal._signal.emit()

    def link(self, slot: Callable[[], None]) -> None:
        self._signal._signal.connect(slot)

    def unlink(self, slot: Callable[[], None] | None = None) -> None:
        if slot is None:
            self._signal._signal.disconnect()
        else:
            self._signal._signal.disconnect(slot)


class OutHollowSignal:
    _signal: _HollowSignal

    def __init__(self, slot: Callable[[], None] | None = None):
        self._signal = _HollowSignal()
        if slot is not None:
            self.link(slot)

    def send(self) -> None:
        self._signal._signal.emit()

    def link(self, slot: Callable[[], None]) -> None:
        self._signal._signal.connect(slot)

    def unlink(self, slot: Callable[[], None] | None = None) -> None:
        if slot is None:
            self._signal._signal.disconnect()
        else:
            self._signal._signal.disconnect(slot)


class InSolidSignal(Generic[_T]):
    _signal: _SolidSignal

    def __init__(self, slot: Callable[[_T], None] | None = None):
        self._signal = _SolidSignal()
        if slot is not None:
            self.link(slot)

    def __call__(self, arg: _T) -> None:
        self._signal._signal.emit(arg)

    def link(self, slot: Callable[[_T], None]) -> None:
        self._signal._signal.connect(slot)

    def unlink(self, slot: Callable[[_T], None] | None = None) -> None:
        if slot is None:
            self._signal._signal.disconnect()
        else:
            self._signal._signal.disconnect(slot)


class OutSolidSignal(Generic[_T]):
    _signal: _SolidSignal

    def __init__(self, slot: Callable[[_T], None] | None = None):
        self._signal = _SolidSignal()
        if slot is not None:
            self.link(slot)

    def send(self, arg: _T) -> None:
        self._signal._signal.emit(arg)

    def link(self, slot: Callable[[_T], None]) -> None:
        self._signal._signal.connect(slot)

    def unlink(self, slot: Callable[[_T], None] | None = None) -> None:
        if slot is None:
            self._signal._signal.disconnect()
        else:
            self._signal._signal.disconnect(slot)



from typing import Generic, TypeVar

from PySide6.QtCore import QObject

_T = TypeVar("_T", bound="QObject")


class QComponent(Generic[_T]):
    __view: _T

    def __init__(self, view: _T):
        self.__view = view

    def view(self) -> _T:
        return self.__view



from clochette.provider.google.dto.GoogleCalendar import GoogleCalendar
from clochette.provider.shared.ui.QOAuth2Panel import QOAuth2Panel


class QGoogleOAuth2Panel(QOAuth2Panel[GoogleCalendar]):
    pass



from dataclasses import dataclass

from clochette.application.service.provider.dto.ProviderDTO import ProviderDTO
from clochette.provider.google.configuration.GoogleSourceConfigurationService import (
    GoogleSourceConfigurationService,
)
from clochette.provider.google.download.GoogleCalendarDownloaderService import (
    GoogleCalendarDownloaderService,
)
from clochette.provider.google.dto.GoogleProviderType import GoogleProviderType
from clochette.provider.google.ui.QGoogleOAuth2Component import QGoogleOAuth2Component


@dataclass
class GoogleProviderFactory:
    """Provider Factory for Google calendars."""

    _download_calendar: GoogleCalendarDownloaderService
    _provider_configuration_mapper: GoogleSourceConfigurationService
    _authentication_component: QGoogleOAuth2Component

    def create(self) -> ProviderDTO:
        """Factory method to create a GoogleProviderDTO with all dependencies."""
        return ProviderDTO(
            download_calendar=self._download_calendar,
            provider_type=GoogleProviderType(),
            provider_configuration_mapper=self._provider_configuration_mapper,
            display_name="Google",
            authentication_component=self._authentication_component,
        )



from PySide6.QtGui import QIntValidator
from PySide6.QtWidgets import QLineEdit


class QIntLineEdit(QLineEdit):

    def __init__(self, placeholder: str | None = None, default_value: int | None = None):
        super().__init__()

        self.setValidator(QIntValidator())

        if placeholder is not None:
            self.setPlaceholderText(placeholder)

        if default_value is not None:
            self.setText(str(default_value))

    def get_value(self) -> int:
        return int(self.text())

    def set_value(self, value: int):
        self.setText(str(value))



from typing import TypeVar, Callable

from PySide6.QtCore import QObject
from reactivex import Observable
from reactivex.abc import DisposableBase

from clochette import log
from clochette.framework.qt.Signal import InSolidSignal
from clochette.framework.rx.Scheduler import scheduler

_T = TypeVar("_T")


class Link:
    """
    Helper class to link an Observable to a handler method via an InSignal for thread safety.

    Automatically handles:
    - Creating an InSignal to make the handler thread-safe
    - Subscribing the Observable to the InSignal
    - Disposing the subscription when the widget is destroyed

    Example:
        Link(
            observable=store.some_observable,
            handler=self._on_value_changed,
            widget=self._view
        )
    """

    _disposable: DisposableBase

    def __init__(
        self,
        observable: Observable[_T],
        handler: Callable[[_T], None],
        widget: QObject,
    ):
        """
        Link an Observable to a handler method via an InSignal.

        Args:
            observable: The Observable to subscribe to
            handler: The method to call when the Observable emits (will be wrapped in InSignal)
            widget: The widget whose destroyed signal will trigger cleanup
        """
        # Create an InSignal to make the handler thread-safe
        in_signal = InSolidSignal(handler)

        # Subscribe the Observable to call the InSignal
        self._disposable = observable.subscribe(
            on_next=lambda value: in_signal(value),
            on_error=lambda e: log.warning("Error in link", exc_info=e),
            scheduler=scheduler,
        )

        # Connect disposal to widget destruction
        widget.destroyed.connect(self._disposable.dispose)



from dataclasses import dataclass
from datetime import datetime

from clochette.domain.entity.CalendarID import CalendarID


@dataclass(frozen=True, eq=True)
class Calendar:
    calendar_id: CalendarID
    last_download: datetime



from abc import ABC, abstractmethod
from typing import Generic, Type, TypeVar

from PySide6.QtWidgets import QWidget

from clochette.domain.entity.CalendarID import CalendarID

_T = TypeVar("_T")
_U = TypeVar("_U")


class IAuthenticationComponent(ABC, Generic[_T, _U]):
    """Generic interface for authentication components"""

    @abstractmethod
    def view(self) -> QWidget:
        """Get the widget view for this component"""
        pass

    @abstractmethod
    def set_values(self, calendar_id: CalendarID, source: _T) -> None:
        """Set values in the component from the source"""
        pass

    @abstractmethod
    def get_values(self) -> _U:
        """Get values from the component"""
        pass

    @abstractmethod
    def get_source(self) -> _T:
        """Get the calendar source from the component values"""
        pass

    @property
    @abstractmethod
    def source_type(self) -> Type[_T]:
        """Get the source type this component handles"""
        pass

    @abstractmethod
    def validate(self) -> bool:
        """Validate the component's values"""
        pass

    @abstractmethod
    def clear(self) -> None:
        """Clear the component's values"""
        pass



from dataclasses import replace

from clochette.domain.entity.CalendarID import CalendarID
from clochette.domain.entity.configuration.CalendarConfiguration import CalendarConfiguration
from clochette.presentation.settings.calendar.dto.CalendarFormValuesDTO import CalendarFormValuesDTO


class CalendarConfigurationFactory:
    def create_from_form_values(self, form_values: CalendarFormValuesDTO) -> CalendarConfiguration:
        """
        Create a new CalendarConfiguration from form values.

        Args:
            form_values: DTO containing all form values from the UI

        Returns:
            New CalendarConfiguration with a generated ID
        """
        return CalendarConfiguration(
            id=CalendarID.new(),
            name=form_values.name,
            source=form_values.source,
            download_interval=form_values.download_interval,
            missed_reminders_past_window=form_values.missed_reminders_past_window,
            force_alarms=form_values.force_alarms,
            force_alarms_dates=form_values.force_alarms_dates,
            default_alarms=form_values.default_alarms,
            default_alarms_dates=form_values.default_alarms_dates,
            http_timeout=form_values.http_timeout,
        )

    def update_from_form_values(
        self, existing: CalendarConfiguration, form_values: CalendarFormValuesDTO
    ) -> CalendarConfiguration:
        """
        Update an existing CalendarConfiguration with new form values.

        Args:
            existing: The existing configuration to update
            form_values: DTO containing updated form values from the UI

        Returns:
            Updated CalendarConfiguration (new instance, since frozen)
        """
        return replace(
            existing,
            name=form_values.name,
            source=form_values.source,
            download_interval=form_values.download_interval,
            missed_reminders_past_window=form_values.missed_reminders_past_window,
            force_alarms=form_values.force_alarms,
            force_alarms_dates=form_values.force_alarms_dates,
            default_alarms=form_values.default_alarms,
            default_alarms_dates=form_values.default_alarms_dates,
            http_timeout=form_values.http_timeout,
        )



from dataclasses import dataclass

from reactivex import operators as ops, Observable

from clochette import log
from clochette.application.repository.CalendarAuthenticationRepository import CalendarAuthenticationRepository
from clochette.application.repository.CalendarRepository import CalendarRepository
from clochette.application.store.CalendarConfigurationStore import CalendarConfigurationStore
from clochette.application.store.GeneralConfigurationStore import GeneralConfigurationStore
from clochette.application.usecase.ScheduleAllCalendarsDownloadUseCase import ScheduleAllCalendarsDownloadUseCase
from clochette.domain.entity.configuration.CalendarConfiguration import CalendarConfiguration
from clochette.framework.rx.Scheduler import scheduler
from clochette.infrastructure.mapper.Mapper import Mapper
from clochette.persist.configuration.ConfigurationReaderService import ConfigurationReaderService
from clochette.persist.configuration.dto.GlobalConfigurationDTO import GlobalConfigurationDTO


@dataclass
class LoadConfigurationUseCase:
    """Use case to load configuration from disk and populate the configuration stores."""

    _configuration_reader_service: ConfigurationReaderService
    _calendar_configuration_store: CalendarConfigurationStore
    _general_configuration_store: GeneralConfigurationStore
    _calendar_authentication_repository: CalendarAuthenticationRepository
    _calendar_repository: CalendarRepository
    _schedule_all_calendars_download_usecase: ScheduleAllCalendarsDownloadUseCase
    _mapper: Mapper

    def load_configuration(self) -> Observable[None]:
        return self._configuration_reader_service.read().pipe(
            ops.do_action(
                on_next=lambda x: self._load_configuration(x),
                on_error=lambda e: log.error(f"Failed to reade configuration", exc_info=e),
            ),
            ops.subscribe_on(scheduler=scheduler),
            ops.map(lambda _: None),
        )

    def _load_configuration(self, global_config_dto: GlobalConfigurationDTO) -> None:
        calendars = [
            self._mapper.map(CalendarConfiguration, calendar_dto) for calendar_dto in global_config_dto.calendars
        ]
        self._general_configuration_store.set_snoozes(global_config_dto.snoozes)
        self._general_configuration_store.set_theme(global_config_dto.theme)
        self._calendar_configuration_store.set_calendars(calendars)
        self._cleanup_calendars(calendars)
        self._update_calendar_repository(calendars)
        self._schedule_all_calendars_download_usecase.schedule_all_calendars(calendars)

    def _cleanup_calendars(self, calendars: list[CalendarConfiguration]):
        ids_to_keep = [calendar.id for calendar in calendars]
        self._calendar_authentication_repository.cleanup_calendars(ids_to_keep)

    def _update_calendar_repository(self, calendars: list[CalendarConfiguration]) -> None:
        """Update calendar repository with the given calendar configurations."""
        if not calendars:
            log.warning("No calendars found in the configuration.")

        log.info(f"Updating calendar repository. Calendars: {calendars}")
        self._calendar_repository.set_calendars(calendars)



from dataclasses import dataclass
from datetime import datetime, date

from clochette.domain.entity.EventID import EventID
from clochette.domain.entity.Trigger import Trigger
from clochette.infrastructure.data_structure.PeekableIterator import PeekableIterator


@dataclass(frozen=True, eq=True)
class CalendarEvent:
    event_id: EventID
    start: date | datetime
    end: date | datetime
    trigger_iterator: PeekableIterator[Trigger]
    raw: str

    def __hash__(self) -> int:
        return hash(self.event_id)

    def __str__(self) -> str:
        return f"CalendarEvent(uid={self.event_id}, start={self.start}, end={self.end}, raw=\n{self.raw})"



from __future__ import annotations
from typing import TypeVar, Generic, Iterator, Iterable, Callable

_T = TypeVar("_T")


class PeekableIterator(Generic[_T]):
    _iterator: Iterator[_T]
    _buffer: list[_T]

    def __init__(self, iterable: Iterable[_T]) -> None:
        self._iterator: Iterator[_T] = iter(iterable)
        self._buffer: list[_T] = []

    def __iter__(self) -> Iterator[_T]:
        return self

    def __next__(self) -> _T:
        if self._buffer:
            return self._buffer.pop()
        return next(self._iterator)

    def _peek(self) -> _T:
        if not self._buffer:
            try:
                trigger = next(self._iterator)
                self._buffer.append(trigger)
            except StopIteration:
                raise StopIteration("No more items to peek")
        return self._buffer[-1]


def take_while(iterator: PeekableIterator[_T], predicate: Callable[[_T], bool]) -> list[_T]:
    result: list[_T] = []
    while True:
        current_value = peek(iterator)
        if current_value is None:
            break

        if predicate(current_value):
            result.append(next(iterator))
        else:
            break

    return result


def peek(iterator: PeekableIterator):
    try:
        return iterator._peek()
    except StopIteration:
        return None



from dataclasses import dataclass

from clochette.application.store.ProviderStore import ProviderStore
from clochette.application.usecase.AddCalendarUseCase import AddCalendarUseCase
from clochette.application.usecase.UpdateCalendarUseCase import UpdateCalendarUseCase
from clochette.domain.entity.configuration.CalendarConfiguration import (
    CalendarConfiguration,
)
from clochette.domain.factory.CalendarConfigurationFactory import (
    CalendarConfigurationFactory,
)
from clochette.framework.qt.Link import Link
from clochette.framework.qt.QComponent import QComponent
from clochette.presentation.settings.calendar.CalendarNavigationModel import (
    CalendarNavigationModel,
)
from clochette.presentation.settings.calendar.CalendarSettingsModel import (
    CalendarSettingsModel,
)
from clochette.presentation.settings.calendar.authentication.AuthenticationConfigurationModel import (
    AuthenticationConfigurationModel,
)
from clochette.presentation.settings.calendar.tabs.QCalendarTabPanelView import (
    QCalendarTabPanelView,
)
from clochette.provider.public.dto.PublicProviderType import PublicProviderType


@dataclass
class QCalendarConfigurationComponent(QComponent[QCalendarTabPanelView]):
    """Coordinator for calendar configuration that wires navigation and use cases"""

    _view: QCalendarTabPanelView
    _provider_store: ProviderStore
    _auth_model: AuthenticationConfigurationModel
    _settings_model: CalendarSettingsModel
    _navigation_model: CalendarNavigationModel
    _add_calendar_usecase: AddCalendarUseCase
    _update_calendar_usecase: UpdateCalendarUseCase
    _calendar_configuration_factory: CalendarConfigurationFactory

    def __post_init__(self):
        super().__init__(self._view)

        # Wire: ProviderStore → View InSignal
        Link(
            observable=self._provider_store.providers,
            handler=self._view.set_providers,
            widget=self._view,
        )

        # Connect OK/Cancel buttons
        self._view.on_ok.link(self._save_calendar)
        self._view.on_cancel.link(lambda: self._navigation_model.show_calendar_list.on_next(True))

        # Wire: Navigation add_calendar → handler (Link wraps in signal for thread-safety)
        Link(
            observable=self._navigation_model.add_calendar,
            handler=lambda _: self._on_add_calendar(),
            widget=self._view,
        )

        # Wire: Navigation edit_calendar → handler (Link wraps in signal for thread-safety)
        Link(
            observable=self._navigation_model.edit_calendar,
            handler=self._on_edit_calendar,
            widget=self._view,
        )

    def _on_add_calendar(self):
        """Handle add calendar navigation event"""
        self._clear()
        self._auth_model.edit_calendar = None
        self._settings_model.edit_calendar = None
        self._view.set_selected_provider_type(PublicProviderType())

    def _on_edit_calendar(self, calendar: CalendarConfiguration):
        """Handle edit calendar navigation event (authentication tab)"""
        self._clear()
        self._auth_model.edit_calendar = calendar
        self._settings_model.edit_calendar = calendar
        self._view.set_configuration(calendar)
        self._view.set_selected_provider_type(calendar.source.provider_type)

    def _clear(self):
        """Clear all form fields"""
        self._view.clear()

    def _save_calendar(self):
        """Save the calendar (add or update)"""
        # Get form values from the view (presentation layer DTO)
        form_values = self._view.get_form_values()

        # Check if we're editing or adding
        existing_calendar = self._auth_model.edit_calendar or self._settings_model.edit_calendar

        if existing_calendar is not None:
            # Update existing calendar using factory
            calendar = self._calendar_configuration_factory.update_from_form_values(existing_calendar, form_values)
            self._update_calendar_usecase.update_calendar(calendar).subscribe()
        else:
            # Add new calendar using factory
            calendar = self._calendar_configuration_factory.create_from_form_values(form_values)
            self._add_calendar_usecase.add_calendar(calendar).subscribe()

        self._navigation_model.show_calendar_list.on_next(True)



import json
from dataclasses import dataclass

from clochette.application.repository.CalendarAuthenticationRepository import CalendarAuthenticationRepository
from clochette.domain.entity.CalendarID import CalendarID
from clochette.provider.basic.dto.BasicAuthDTO import BasicAuthDTO


@dataclass
class BasicAuthenticationService:
    _calendar_authentication_repository: CalendarAuthenticationRepository

    def store_auth(self, calendar_id: CalendarID, auth: BasicAuthDTO) -> None:
        self._calendar_authentication_repository.store_value(calendar_id, json.dumps(auth, default=vars))

    def retrieve_auth(self, calendar_id: CalendarID) -> BasicAuthDTO | None:
        value = self._calendar_authentication_repository.get_value(calendar_id)
        if value:
            json_value = json.loads(value)
            return BasicAuthDTO(json_value.get("username"), json_value.get("password"), json_value.get("cancelled"))

        return None



from pathlib import Path

import platformdirs


class PlatformConfigurationService:
    APP_NAME = "clochette"
    AUTHOR = "sketyl"
    CONFIG_FILE = "clochette.config"
    LOGGING_FILE = "logging.config"

    def __init__(self) -> None:
        self.application_configuration_path().mkdir(parents=True, exist_ok=True)
        self.application_data_path().mkdir(parents=True, exist_ok=True)

    def application_configuration_path(self) -> Path:
        config = platformdirs.user_config_dir(
            PlatformConfigurationService.APP_NAME, PlatformConfigurationService.AUTHOR
        )
        return Path(config)

    def application_data_path(self) -> Path:
        data = platformdirs.user_data_dir(PlatformConfigurationService.APP_NAME, PlatformConfigurationService.AUTHOR)
        return Path(data)

    def application_configuration_file(self) -> Path:
        return self.application_configuration_path() / PlatformConfigurationService.CONFIG_FILE

    def logging_configuration_file(self) -> Path:
        return self.application_configuration_path() / PlatformConfigurationService.LOGGING_FILE



from typing import TypeVar, Generic

from PySide6.QtWidgets import QGroupBox, QFormLayout, QPushButton, QLabel, QHBoxLayout

from clochette import log
from clochette.framework.qt.Signal import OutHollowSignal, InHollowSignal, InSolidSignal
from clochette.presentation.widget.QCheckBoxForm import QCheckBoxForm, CheckBoxItem
from clochette.presentation.widget.QWidgetLayout import QWidgetLayout
from clochette.provider.shared.dto.AuthCalendar import AuthCalendar

_T = TypeVar("_T", bound=AuthCalendar)


class QOAuth2Panel(QGroupBox, Generic[_T]):
    on_authenticate: OutHollowSignal

    set_authentication_failed: InSolidSignal[Exception]
    set_authentication_success: InHollowSignal
    clear: InHollowSignal
    set_error: InSolidSignal[str]
    set_selected_calendars: InSolidSignal[dict[_T, bool]]

    _auth_button: QPushButton
    _checkbox_form: QCheckBoxForm[_T]
    _error_label: QLabel

    def __init__(self):
        super().__init__("")

        self.on_authenticate = OutHollowSignal()
        self.set_authentication_failed = InSolidSignal(self._on_authentication_failed)
        self.set_authentication_success = InHollowSignal(self._on_authentication_success)
        self.clear = InHollowSignal(self._on_clear)
        self.set_error = InSolidSignal(self._on_set_error)
        self.set_selected_calendars = InSolidSignal(self._on_set_selected_calendars)

        layout = QFormLayout()
        self.setLayout(layout)

        button_layout = QWidgetLayout(QHBoxLayout())
        button_layout.add_stretch()

        self._auth_button = QPushButton(self.tr("Authenticate"))
        self._auth_button.clicked.connect(self._authenticate)
        button_layout.add_widget(self._auth_button)

        self._checkbox_form = QCheckBoxForm(self.tr("Calendars"))
        layout.addRow(self._checkbox_form)

        layout.addRow(button_layout)

        self._error_label = QLabel()
        self._error_label.setWordWrap(True)
        layout.addRow(self._error_label)

    def _authenticate(self):
        self._auth_button.setText(self.tr("Authenticating..."))
        self._auth_button.setEnabled(False)
        self.on_authenticate.send()

    def _on_authentication_failed(self, e: Exception):
        self._error_label.setStyleSheet("color: red;")
        self._error_label.setText(self.tr("Failed to authenticate: %s") % e)
        self._auth_button.setText(self.tr("Authenticate"))
        self._auth_button.setEnabled(True)
        log.error("Failed to authenticate", exc_info=e)

    def _on_authentication_success(self):
        self._error_label.setStyleSheet("color: green;")
        self._error_label.setText(self.tr("Authentication successful"))
        self._auth_button.setText(self.tr("Authenticate"))
        self._auth_button.setEnabled(True)

    def _on_clear(self):
        self._auth_button.setText(self.tr("Authenticate"))
        self._error_label.setText("")

    def _on_set_error(self, error: str) -> None:
        self._error_label.setStyleSheet("color: red;")
        self._error_label.setText(error)

    def _on_set_selected_calendars(self, calendars: dict[_T, bool]) -> None:
        items = {CheckBoxItem(cal.summary, cal): checked for cal, checked in calendars.items()}
        self._checkbox_form.set_checkbox_states(items)

    @property
    def selected_calendars(self) -> dict[CheckBoxItem[_T], bool]:
        return self._checkbox_form.get_checkbox_states()



from PySide6.QtGui import Qt
from PySide6.QtWidgets import QWidget, QVBoxLayout, QPushButton, QHBoxLayout

from clochette.application.store.CalendarConfigurationStore import CalendarConfigurationStore
from clochette.application.usecase.DeleteCalendarUseCase import DeleteCalendarUseCase
from clochette.domain.entity.configuration.CalendarConfiguration import CalendarConfiguration
from clochette.framework.qt.Link import Link
from clochette.presentation.settings.calendar.CalendarNavigationModel import CalendarNavigationModel
from clochette.presentation.widget.QList import QList, ListItem


class QCalendarList(QWidget):

    _list_calendar: QList[CalendarConfiguration]
    _calendar_configuration_store: CalendarConfigurationStore
    _calendar_navigation_model: CalendarNavigationModel
    _delete_calendar_usecase: DeleteCalendarUseCase

    def __init__(
        self,
        calendar_configuration_store: CalendarConfigurationStore,
        calendar_navigation_model: CalendarNavigationModel,
        delete_calendar_usecase: DeleteCalendarUseCase,
    ) -> None:
        super().__init__()

        self._calendar_configuration_store = calendar_configuration_store
        self._calendar_navigation_model = calendar_navigation_model
        self._delete_calendar_usecase = delete_calendar_usecase

        main_layout = QHBoxLayout()
        self.setLayout(main_layout)

        self._list_calendar = QList()

        # Create the buttons layout
        buttons_layout = QVBoxLayout()
        buttons_layout.setAlignment(Qt.AlignmentFlag.AlignTop)

        # Create the Add button
        add_button = QPushButton(self.tr("Add"))
        add_button.clicked.connect(lambda: calendar_navigation_model.add_calendar.on_next(True))
        buttons_layout.addWidget(add_button)

        # Create the Edit button
        edit_button = QPushButton(self.tr("Edit"))
        edit_button.clicked.connect(self._edit_calendar)
        buttons_layout.addWidget(edit_button)

        # Create the Delete button
        delete_button = QPushButton(self.tr("Delete"))
        delete_button.clicked.connect(self._delete_calendar)
        buttons_layout.addWidget(delete_button)

        # Add the list widget and the buttons layout to the main layout
        main_layout.addWidget(self._list_calendar)
        main_layout.addLayout(buttons_layout)
        main_layout.addStretch()

        Link(
            observable=calendar_configuration_store.calendars,
            handler=self._set_calendars,
            widget=self,
        )

    def _set_calendars(self, calendars: list[CalendarConfiguration]):
        self._list_calendar.clear()
        for calendar in calendars:
            self._list_calendar.add(ListItem(calendar.name, calendar))

        calendar = self._list_calendar.get_selected_item()
        if calendars and calendar is None:
            self._list_calendar.setCurrentRow(0)

    def selected_calendar(self) -> CalendarConfiguration | None:
        return self._list_calendar.get_selected_item()

    def _edit_calendar(self):
        calendar = self._list_calendar.get_selected_item()
        if calendar is not None:
            self._calendar_navigation_model.edit_calendar.on_next(calendar)

    def _delete_calendar(self):
        calendar = self._list_calendar.get_selected_item()
        if calendar is not None:
            self._delete_calendar_usecase.delete_calendar(calendar).subscribe()



from dataclasses import dataclass, field

from clochette.application.service.provider.dto.ProviderType import ProviderType
from clochette.domain.entity.ISourceCalendar import ISourceCalendar
from clochette.provider.google.dto.GoogleCalendar import GoogleCalendar
from clochette.provider.google.dto.GoogleProviderType import GoogleProviderType


@dataclass(frozen=True, eq=True)
class GoogleSource(ISourceCalendar):
    calendars: dict[GoogleCalendar, bool]
    provider_type: ProviderType = field(default_factory=GoogleProviderType, init=False, compare=False)

    @property
    def selected_calendars(self) -> list[GoogleCalendar]:
        return [cal for cal, checked in self.calendars.items() if checked]



from dataclasses import dataclass

from reactivex import operators as ops, Observable

from clochette import log
from clochette.application.usecase.ScheduleCalendarDownloadUseCase import ScheduleCalendarDownloadUseCase
from clochette.application.usecase.UnscheduleCalendarDownloadUseCase import UnscheduleCalendarDownloadUseCase
from clochette.domain.entity.configuration.CalendarConfiguration import CalendarConfiguration
from clochette.framework.rx.Scheduler import scheduler


@dataclass
class RescheduleCalendarDownloadUseCase:
    """Use case to reschedule downloads for a calendar."""

    _unschedule_calendar_download_usecase: UnscheduleCalendarDownloadUseCase
    _schedule_calendar_download_usecase: ScheduleCalendarDownloadUseCase

    def reschedule_calendar(self, calendar: CalendarConfiguration) -> Observable[None]:
        """Reschedule downloads for a calendar (unschedule old, schedule new)."""
        log.info(f"Rescheduling downloads for calendar: {calendar.name}")

        # Unschedule the old task
        return self._unschedule_calendar_download_usecase.unschedule_calendar(calendar).pipe(
            ops.map(lambda _: None),
            ops.do_action(
                on_next=lambda _: self._schedule_calendar_download_usecase.schedule_calendar(calendar),
                on_error=lambda e: log.error(
                    f"Failed to reschedule downloads for calendar named: {calendar.name}", exc_info=e
                ),
            ),
            ops.subscribe_on(scheduler=scheduler),
        )



import keyring

from clochette import log


class KeyringService:
    _service_name: str

    def __init__(self) -> None:
        self._service_name = "clochette"

    def store_value(self, key: str, value: str) -> None:
        log.debug(f"Storing value in keyring, prefix: key: {key}")
        keyring.set_password(self._service_name, key, value)

    def get_value(self, key: str) -> str | None:
        log.debug(f"Retrieving value from keyring, key: {key}")
        return keyring.get_password(self._service_name, key)

    def delete_value(self, key: str) -> None:
        log.debug(f"Deleting value from keyring, key: {key}")
        try:
            keyring.delete_password(self._service_name, key)
        except Exception:
            log.warning(f"Deleting keyring value which doesn't exists: {key}")



from dataclasses import dataclass
from typing import Type, override

from clochette.application.service.provider.dto.IAuthenticationComponent import IAuthenticationComponent
from clochette.domain.entity.CalendarID import CalendarID
from clochette.framework.qt.QComponent import QComponent
from clochette.infrastructure.http_.client.HttpService import HttpService
from clochette.provider.public.configuration.PublicURLSource import PublicURLSource
from clochette.provider.public.ui.QPublicAuthPanel import QPublicAuthPanel


@dataclass(frozen=True)
class PublicValues:
    url: str


@dataclass
class QPublicAuthComponent(QComponent[QPublicAuthPanel], IAuthenticationComponent[PublicURLSource, PublicValues]):
    _view: QPublicAuthPanel
    _http_service: HttpService

    def __post_init__(self):
        super().__init__(self._view)
        self._view.on_authenticate.link(self._public_auth)

    def _public_auth(self):
        try:
            self._http_service.get(self._view.url)
            self._view.set_authentication_success()
        except Exception as e:
            self._view.set_authentication_failed(e)

    @override
    def clear(self):
        self._view.clear()

    @override
    def get_values(self) -> PublicValues:
        return PublicValues(self._view.url)

    @override
    def get_source(self) -> PublicURLSource:
        return PublicURLSource(self._view.url)

    @property
    @override
    def source_type(self) -> Type[PublicURLSource]:
        return PublicURLSource

    @override
    def set_values(self, calendar_id: CalendarID, source: PublicURLSource) -> None:
        self._view.set_url(source.url)

    @override
    def validate(self) -> bool:
        if not self._view.url:
            self._view.set_error("URL cannot be empty")
            return False
        return True



from dataclasses import dataclass
from datetime import timedelta


from clochette import log
from clochette.application.store.SnoozeScheduleStore import SnoozeScheduleStore
from clochette.application.usecase.CheckForSnoozeUseCase import CheckForSnoozeUseCase
from clochette.infrastructure.schedule.SchedulerService import SchedulerService


@dataclass
class ScheduleSnoozeReminderUseCase:
    """Use case to schedule snooze reminder checks."""

    _snooze_schedule_store: SnoozeScheduleStore
    _check_for_snooze_use_case: CheckForSnoozeUseCase
    _scheduler_service: SchedulerService

    def schedule_snooze_reminders(self) -> None:
        """Schedule periodic snooze reminder checks."""
        log.info("Scheduling snooze reminder checks")

        task_id = self._scheduler_service.schedule_at_interval(
            timedelta(seconds=10),
            self._check_for_snooze_use_case.check_for_snoozes,
        )

        self._snooze_schedule_store.set_schedule_id(task_id)
        log.debug(f"Scheduled snooze reminder checks, task_id: {task_id}")



from dataclasses import dataclass
from datetime import datetime, UTC

from peewee import AutoField, TextField

from clochette.persist.database.field.DateTimeField import DateTimeField
from clochette.persist.database.model.BaseModel import BaseModel


class DatabaseVersion(BaseModel):
    id = AutoField()
    version = TextField(null=False)
    applied_on = DateTimeField(null=False, default=datetime.now(tz=UTC))

    class Meta:  # pyright: ignore [reportIncompatibleVariableOverride]
        table_name = "database_version"


@dataclass(frozen=True, eq=True)
class DatabaseVersionDTO:
    id: int
    version: str
    applied_on: datetime



from PySide6.QtCore import QSize
from PySide6.QtGui import QPixmap, Qt
from PySide6.QtWidgets import QLabel


class QWidgetIcon(QLabel):
    def __init__(self, image_path: str, size: QSize) -> None:
        super().__init__()
        pixmap = QPixmap(image_path)
        scaled_pixmap = pixmap.scaled(size, Qt.AspectRatioMode.KeepAspectRatio)
        self.setPixmap(scaled_pixmap)
        self.setAlignment(Qt.AlignmentFlag.AlignCenter)



from datetime import datetime

from clochette.domain.entity.EventID import EventID
from clochette.domain.entity.Occurrence import Occurrence
from clochette.domain.entity.Snooze import Snooze
from clochette.infrastructure.data_structure.AtomicDict import AtomicDict


class SnoozeQuery:
    _snoozes: AtomicDict[EventID, Snooze]

    def __init__(self):
        self._snoozes = AtomicDict({})

    def before(self, now: datetime) -> list[Occurrence]:
        occurrences = []

        for _, snooze in self._snoozes.items():
            trigger = snooze.trigger

            if trigger.trigger > now:
                continue

            occurrences.append(Occurrence(snooze.event_id, trigger))

        return occurrences

    def add(self, snooze: Snooze) -> None:
        self._snoozes[snooze.event_id] = snooze



from dataclasses import dataclass
from typing import Type, override

from clochette.application.service.provider.dto.IAuthenticationComponent import IAuthenticationComponent
from clochette.domain.entity.CalendarID import CalendarID
from clochette.domain.entity.configuration.HttpTimeout import HttpTimeout
from clochette.framework.qt.QComponent import QComponent
from clochette.presentation.settings.calendar.authentication.AuthenticationConfigurationModel import (
    AuthenticationConfigurationModel,
)
from clochette.provider.basic.configuration.BasicAuthURLSource import BasicAuthURLSource
from clochette.provider.basic.download.BasicAuthService import BasicAuthService, Credentials
from clochette.provider.basic.keyring_.BasicAuthenticationService import BasicAuthenticationService
from clochette.provider.basic.ui.QBasicAuthView import QBasicAuthView


@dataclass(frozen=True)
class BasicValues:
    url: str
    username: str
    password: str


@dataclass
class QBasicAuthComponent(QComponent[QBasicAuthView], IAuthenticationComponent[BasicAuthURLSource, BasicValues]):
    _view: QBasicAuthView
    _model: AuthenticationConfigurationModel
    _basic_auth_service: BasicAuthService
    _basic_authentication_service: BasicAuthenticationService

    def __post_init__(self):
        super().__init__(self._view)
        self._view.on_authenticate.link(self._basic_auth)

    def _basic_auth(self):
        try:
            calendar_id = self._model.calendar_id
            self._basic_auth_service.download(
                calendar_id,
                BasicAuthURLSource(self._view.url),
                HttpTimeout.default(),
                Credentials(self._view.username, self._view.password),
            )
            self._view.set_authentication_success()
        except Exception as e:
            self._view.set_authentication_failed(e)

    @override
    def clear(self):
        self._view.clear()

    @override
    def get_values(self) -> BasicValues:
        return BasicValues(self._view.url, self._view.username, self._view.password)

    @override
    def get_source(self) -> BasicAuthURLSource:
        return BasicAuthURLSource(self._view.url)

    @property
    @override
    def source_type(self) -> Type[BasicAuthURLSource]:
        return BasicAuthURLSource

    @override
    def set_values(self, calendar_id: CalendarID, source: BasicAuthURLSource) -> None:
        self._view.set_url(source.url)

        # Retrieve credentials from keyring
        auth = self._basic_authentication_service.retrieve_auth(calendar_id)
        if auth is None:
            self._view.set_username("")
            self._view.set_password("")
        else:
            self._view.set_username(auth.username)
            self._view.set_password(auth.password)

    @override
    def validate(self) -> bool:
        if not self._view.url:
            self._view.set_error("URL cannot be empty")
            return False
        return True



from datetime import datetime

from peewee import Field


class DateTimeField(Field):
    field_type = "text"

    def db_value(self, value: datetime) -> str:
        return value.isoformat()

    def python_value(self, value: str) -> datetime:
        return datetime.fromisoformat(value)



import inspect
from typing import Type, get_type_hints, TypeVar, Any, get_origin

from clochette import log

_T = TypeVar("_T")


class Container:
    _container: dict[Type, Any]

    def __init__(self):
        self._container = {}

    def instantiate(self, cls: Type[_T]) -> _T:

        instance = self._container.get(cls, None)
        if instance is not None:
            return instance

        cls = self._get_class_without_generic(cls)
        if inspect.isclass(cls):
            if inspect.isabstract(cls):
                raise Exception(f"Cannot instantiate abstract class: {cls}")
        else:
            raise Exception(f"Cannot instantiate type: {cls}")

        # Get the __init__ method of the class
        init_method = cls.__init__

        # Get the signature of the __init__ method
        signature = inspect.signature(init_method)

        # Extract the parameters
        parameters = signature.parameters

        # Get the type hints for the parameters
        type_hints = get_type_hints(init_method)

        # Create a dictionary to store parameter names, their types, and instances
        kwargs = {}

        for param_name, param in parameters.items():
            # Skip 'self' parameter
            if param_name == "self":
                continue

            # Skip *args and **kwargs based on parameter kind
            if param.kind == param.VAR_POSITIONAL or param.kind == param.VAR_KEYWORD:
                continue

            # skip parameter with default values
            if param.default is not inspect.Parameter.empty:
                continue

            # Get the type of the parameter from type hints
            param_type = type_hints.get(param_name, inspect.Parameter.empty)

            param_type = self._get_class_without_generic(param_type)

            if inspect.isclass(param_type):
                if inspect.isabstract(param_type):
                    raise Exception(f"Cannot instantiate abstract class: {param_type}, class {cls}")
            else:
                raise Exception(f"Cannot instantiate type: {param_type}, class {cls}")

            kwargs[param.name] = self.instantiate(param_type)

        try:
            instance = cls(**kwargs)
        except Exception:
            log.error("You might want to make your class mutable if you encounter this issue", exc_info=True)
            raise

        self._container[cls] = instance
        return instance

    def inject(self, type: Type[_T], instance: _T):
        self._container[type] = instance

    def _get_class_without_generic(self, param_type: Any):
        origin = get_origin(param_type)
        if origin is not None:
            return origin

        return param_type



from dataclasses import dataclass

from clochette import log
from clochette.domain.entity.Snooze import Snooze
from clochette.domain.entity.CalendarID import CalendarID
from clochette.domain.entity.EventID import EventID
from clochette.domain.entity.EventUID import EventUID
from clochette.domain.entity.Trigger import Trigger
from clochette.persist.database.dao.SnoozeDAO import SnoozeDAO
from clochette.persist.database.model.Snooze import SnoozeDTO


@dataclass
class SnoozeRepository:
    _snooze_dao: SnoozeDAO

    def delete_snooze(self, id: int) -> None:
        log.debug(f"Deleting snooze with ID: {id}")
        self._snooze_dao.delete_snooze(id)

    def delete_snoozes(self, event_ids: list[EventID]):
        for event_id in event_ids:
            self._snooze_dao.delete_snooze_by_event_id(event_id.event_uid.id, event_id.calendar_id.id)

    def add_snooze(self, snooze: Snooze) -> Snooze:
        log.debug(f"Saving snooze: {snooze}")
        id = self._snooze_dao.add_snooze(
            snooze.event_id.event_uid.id, snooze.event_id.calendar_id.id, snooze.trigger.trigger, snooze.trigger.start
        )
        return Snooze(id, snooze.event_id, snooze.trigger)

    def get_snoozes_by_calendar(self, calendar_id: CalendarID) -> list[Snooze]:
        log.debug("Retrieving snoozes")
        snoozes = self._snooze_dao.get_snoozes_by_calendar(calendar_id.id)
        return list(map(self._to_snooze, snoozes))

    def _to_snooze(self, snooze: SnoozeDTO) -> Snooze:
        return Snooze(
            snooze.id,
            EventID(EventUID(snooze.event_uid), CalendarID(snooze.calendar_id)),
            Trigger(snooze.trigger, snooze.trigger_start),
        )



from PySide6.QtWidgets import QWidget, QHBoxLayout, QPushButton

from clochette.framework.qt.Signal import OutHollowSignal


class QOkCancelButton(QWidget):
    on_ok_clicked: OutHollowSignal
    on_cancel_clicked: OutHollowSignal

    _ok_button: QPushButton
    _cancel_button: QPushButton

    def __init__(self, ok_title: str | None = None):
        super().__init__()

        self.on_ok_clicked = OutHollowSignal()
        self.on_cancel_clicked = OutHollowSignal()

        main_layout = QHBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(main_layout)

        self._ok_button = QPushButton(ok_title or self.tr("OK"))
        self._ok_button.clicked.connect(self.on_ok_clicked.send)

        self._cancel_button = QPushButton(self.tr("Cancel"))
        self._cancel_button.clicked.connect(self.on_cancel_clicked.send)

        main_layout.addStretch()
        main_layout.addWidget(self._cancel_button)
        main_layout.addWidget(self._ok_button)

    def set_enabled_ok(self, enable: bool):
        self._ok_button.setEnabled(enable)

    def set_text_ok(self, text: str):
        self._ok_button.setText(text)



from dataclasses import dataclass

from reactivex import Observable, from_callable

from clochette import log
from clochette.application.service.provider.dto.IDownloadCalendar import IDownloadCalendar
from clochette.domain.entity.configuration.CalendarConfiguration import CalendarConfiguration
from clochette.domain.entity.configuration.HttpTimeout import HttpTimeout
from clochette.framework.rx.Scheduler import scheduler
from clochette.provider.basic.configuration.BasicAuthURLSource import BasicAuthURLSource
from clochette.provider.basic.download.BasicAuthService import BasicAuthService


@dataclass
class BasicAuthCalendarDownloaderService(IDownloadCalendar):
    _basic_auth_service: BasicAuthService

    def download(self, calendar_configuration: CalendarConfiguration) -> Observable[str | None]:
        return from_callable(
            lambda: self._download(
                calendar_configuration,
                calendar_configuration.http_timeout,
                calendar_configuration.name,
            ),
            scheduler=scheduler,
        )

    def _download(
        self,
        calendar_configuration: CalendarConfiguration,
        http_timeout: HttpTimeout,
        calendar_name: str,
    ) -> str | None:
        source = calendar_configuration.source
        if not isinstance(source, BasicAuthURLSource):
            raise TypeError(f"Expected BasicAuthURLSource, got {type(source)}")

        log.info(f"Downloading calendar: {calendar_name}")

        return self._basic_auth_service.download(calendar_configuration.id, source, http_timeout)



from dataclasses import dataclass

from clochette.domain.entity.EventDetails import EventDetails
from clochette.domain.entity.Occurrence import Occurrence
from clochette.framework.qt.QComponent import QComponent
from clochette.framework.qt.Signal import InSolidSignal
from clochette.presentation.reminder_window.QReminderListView import QReminderListView


@dataclass
class QReminderListComponent(QComponent[QReminderListView]):
    """Component that manages the reminder list view with business logic."""

    _view: QReminderListView
    update_occurrences: InSolidSignal[list[tuple[Occurrence, EventDetails]]]

    def __post_init__(self) -> None:
        super().__init__(self._view)
        self.update_occurrences = InSolidSignal(self._on_update_occurrences)

    def _on_update_occurrences(self, occurrences: list[tuple[Occurrence, EventDetails]]) -> None:
        """Called when occurrences are updated - forwards to view in thread-safe manner."""
        self._view.set_occurrences(occurrences)



from PySide6.QtCore import QFile, QIODeviceBase


def read_file(file: QFile) -> bytes:
    if file.exists():
        if file.open(QIODeviceBase.OpenModeFlag.ReadOnly):
            return bytes(file.readAll().data())
    raise Exception(f"Cannot open file, make sure it exists or the system has permissions: {file}")



from dataclasses import dataclass

from clochette import log
from clochette.persist.configuration.ConfigurationWriterService import ConfigurationWriterService
from clochette.persist.configuration.PlatformConfigurationService import PlatformConfigurationService
from clochette.persist.configuration.dto.GlobalConfigurationDTO import GlobalConfigurationDTO


@dataclass
class ConfigurationMigrationService:
    _configuration_writer_service: ConfigurationWriterService
    _platform_configuration_service: PlatformConfigurationService

    def init_application_configuration(self) -> None:
        configuration_file = self._platform_configuration_service.application_configuration_file()

        if not configuration_file.exists():
            global_configuration = GlobalConfigurationDTO.default()
            self._configuration_writer_service.write(global_configuration).run()

    def init_logging_configuration(self) -> None:
        configuration_file = self._platform_configuration_service.logging_configuration_file()

        if not configuration_file.exists():
            configuration_content = """[Config]
# Set the desired log level. Default value is WARNING, possible values are: CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET
# Use NOTSET if you want to completely disable the logs 
log.level=WARNING

# Enable the console logs. Default value is FALSE. Possible values are TRUE, FALSE
log.console.enabled=FALSE
"""

            configuration_file.write_text(configuration_content)
            log.info(f"Logging configuration file doesn't exist, creating it: {configuration_file}")
            # print in case logging hasn't been initialized yet
            print(f"Logging configuration file doesn't exist, creating it: {configuration_file}")



from datetime import datetime

from clochette.domain.entity.Occurrence import Occurrence
from clochette.domain.entity.Snooze import Snooze
from clochette.domain.entity.SnoozeQuery import SnoozeQuery
from clochette import log


class SnoozeModel:
    _snooze_query: SnoozeQuery

    def __init__(self) -> None:
        self._snooze_query = SnoozeQuery()

    def add_snooze(self, snooze: Snooze) -> None:
        log.debug(f"Adding new snooze to the model: {snooze}")
        self._snooze_query.add(snooze)

    def query_occurrences(self, now: datetime) -> list[Occurrence]:
        log.debug(f"Querying snoozed occurrences up to: {now}")

        return self._snooze_query.before(now)



from dataclasses import dataclass

from reactivex import operators as ops, Observable, combine_latest

from clochette import log
from clochette.application.store.CalendarConfigurationStore import CalendarConfigurationStore
from clochette.application.usecase.PersistConfigurationUseCase import PersistConfigurationUseCase
from clochette.application.usecase.RescheduleCalendarDownloadUseCase import RescheduleCalendarDownloadUseCase
from clochette.domain.entity.configuration.CalendarConfiguration import CalendarConfiguration
from clochette.framework.rx.Scheduler import scheduler


@dataclass
class UpdateCalendarUseCase:
    """Use case to update an existing calendar in the configuration store."""

    _calendar_configuration_store: CalendarConfigurationStore
    _persist_configuration_usecase: PersistConfigurationUseCase
    _reschedule_calendar_download_usecase: RescheduleCalendarDownloadUseCase

    def update_calendar(self, calendar: CalendarConfiguration) -> Observable[None]:
        """Update an existing calendar in the configuration store and persist to disk."""

        log.info(f"Updating calendar: {calendar}")

        def update_store(calendars: list[CalendarConfiguration]) -> Observable[None]:
            # Replace the calendar in the list
            updated_calendars = [calendar if cal.id == calendar.id else cal for cal in calendars]
            self._calendar_configuration_store.set_calendars(updated_calendars)
            self._calendar_configuration_store.update_calendar(calendar)

            return combine_latest(
                self._reschedule_calendar_download_usecase.reschedule_calendar(calendar),
                self._persist_configuration_usecase.persist_configuration(),
            ).pipe(
                ops.map(lambda _: None),
            )

        return self._calendar_configuration_store.calendars.pipe(
            ops.take(1),
            ops.flat_map(lambda x: update_store(x)),
            ops.do_action(
                on_error=lambda e: log.error(f"Failed to update calendar named: {calendar.name}", exc_info=e),
            ),
            ops.subscribe_on(scheduler=scheduler),
        )



from dataclasses import dataclass
from datetime import datetime
from typing import Generic

from clochette.infrastructure.clock.DateTimeUtils import DateTimeUtils
from clochette.infrastructure.clock.Generics import DateOrDatetimeType


@dataclass(frozen=True, eq=True)
class Trigger(Generic[DateOrDatetimeType]):
    trigger: datetime  # datetime when to notify of an event, must have a TZ
    start: DateOrDatetimeType  # start date/time of the event

    def __post_init__(self):
        if not DateTimeUtils.is_datetime(self.trigger):
            raise ValueError(f"trigger_utc should be of type datetime, {self.trigger}")

        if not self.trigger.tzinfo:
            raise ValueError(f"trigger doesn't have a timezone, {self.trigger}")



from dataclasses import dataclass, field

from clochette.application.service.provider.dto.ProviderType import ProviderType
from clochette.domain.entity.ISourceCalendar import ISourceCalendar
from clochette.provider.microsoft.dto.MicrosoftCalendar import MicrosoftCalendar
from clochette.provider.microsoft.dto.MicrosoftProviderType import MicrosoftProviderType


@dataclass(frozen=True, eq=True)
class MicrosoftSource(ISourceCalendar):
    calendars: dict[MicrosoftCalendar, bool]
    provider_type: ProviderType = field(default_factory=MicrosoftProviderType, init=False, compare=False)

    @property
    def selected_calendars(self) -> list[MicrosoftCalendar]:
        return [cal for cal, checked in self.calendars.items() if checked]



from datetime import datetime, date

from peewee import Field


class DateOrTimeField(Field):
    field_type = "text"

    def db_value(self, value: date | datetime | None) -> str | None:
        if not value:
            return None
        return value.isoformat()

    def python_value(self, value: str | None) -> datetime | date | None:
        if not value:
            return None
        if "T" in value:
            return datetime.fromisoformat(value)
        else:
            return date.fromisoformat(value)



from dataclasses import fields, is_dataclass
from typing import TypeVar, Type, Any

_TOut = TypeVar("_TOut")


class Mapper:
    """Generic object mapper that maps fields from one type to another."""

    def map(
        self,
        type_out: Type[_TOut],
        obj: object,
        **kwargs: Any,
    ) -> _TOut:
        # Ensure type_out is a dataclass
        if not is_dataclass(type_out):
            raise TypeError(f"type_out must be a dataclass, got {type_out}")

        # Get all fields from the target type
        target_fields = fields(type_out)  # type: ignore[arg-type]

        # Build constructor arguments
        constructor_args = {}

        for field in target_fields:
            field_name = field.name

            # Try to get value from source object
            if hasattr(obj, field_name):
                constructor_args[field_name] = getattr(obj, field_name)
            # Otherwise use kwargs
            elif field_name in kwargs:
                constructor_args[field_name] = kwargs[field_name]
            else:
                raise ValueError(
                    f"Field '{field_name}' is missing in source object and not provided in kwargs. "
                    f"Available kwargs: {list(kwargs.keys())}"
                )

        # Create and return the new instance
        return type_out(**constructor_args)



from dataclasses import dataclass

from clochette import log
from clochette.application.service.download.DownloaderService import DownloaderService
from clochette.application.store.DownloadScheduleStore import DownloadScheduleStore
from clochette.domain.entity.configuration.CalendarConfiguration import CalendarConfiguration
from clochette.infrastructure.schedule.SchedulerService import SchedulerService


@dataclass
class ScheduleCalendarDownloadUseCase:
    """Use case to schedule downloads for a single calendar."""

    _download_schedule_store: DownloadScheduleStore
    _scheduler_service: SchedulerService
    _downloader_service: DownloaderService

    def schedule_calendar(self, calendar: CalendarConfiguration) -> None:
        """Schedule downloads for a calendar."""
        log.info(f"Scheduling downloads for calendar: {calendar.name}")

        task_id = self._scheduler_service.schedule_at_interval(
            calendar.download_interval,
            self._downloader_service.download_and_notify(calendar),
        )

        self._download_schedule_store.set_schedule_id(calendar.id, task_id)
        log.debug(f"Scheduled calendar download: {calendar.name}, task_id: {task_id}")



from datetime import datetime

from clochette.infrastructure.data_structure.AtomicDict import AtomicDict
from clochette.domain.entity.CalendarEvent import CalendarEvent
from clochette.domain.entity.CalendarEventQuery import CalendarEventQuery
from clochette.domain.entity.CalendarID import CalendarID
from clochette.domain.entity.EventID import EventID
from clochette.domain.entity.Occurrence import Occurrence
from clochette import log


class EventModel:
    _calendar_queries: AtomicDict[CalendarID, CalendarEventQuery]

    def __init__(self) -> None:
        self._calendar_queries = AtomicDict({})

    def add_query(self, calendar_id: CalendarID, query: CalendarEventQuery) -> None:
        log.debug(f"Adding new model for calendar: {calendar_id}")
        self._calendar_queries[calendar_id] = query

    def query_occurrences(self, now: datetime) -> list[Occurrence]:
        log.debug(f"Querying occurrences up to: {now}")
        query_results: list[Occurrence] = []
        for _, query in self._calendar_queries.items():
            occurrences = query.before(now)
            query_results += occurrences

        return query_results

    def get_event(self, event_id: EventID) -> CalendarEvent | None:
        calendar = self._calendar_queries[event_id.calendar_id]

        if not calendar:
            log.error(f"No calendar found with id: {event_id.calendar_id}")
            return None

        event = calendar.get_event(event_id)

        if event:
            log.debug(f"Found event with id: {event_id}")
        else:
            log.error(f"No event found with id: {event_id}")

        return event



from http.server import BaseHTTPRequestHandler, HTTPServer
from socketserver import BaseRequestHandler
from typing import Any, Callable, TypeAlias

HttpRequestHandler: TypeAlias = Callable[[BaseHTTPRequestHandler], None]
HttpRequestHandlerBuilder: TypeAlias = Callable[[Any, Any, HTTPServer], BaseRequestHandler]


class RequestHandlerFactory:
    _request_handler: HttpRequestHandler

    def __init__(self, request_handler: HttpRequestHandler):
        self._request_handler = request_handler

    def __call__(self, request: Any, client_address: Any, server: HTTPServer) -> BaseHTTPRequestHandler:
        class CustomRequestHandler(BaseHTTPRequestHandler):
            _request_handler: HttpRequestHandler

            def log_message(self, format: str, *args: Any) -> None:
                # Override this method to disable logging
                pass

            def __init__(self, *args: Any, request_handler: HttpRequestHandler, **kwargs: Any) -> None:
                self._request_handler = request_handler
                super().__init__(*args, **kwargs)

            def do_GET(self) -> None:
                return self._request_handler(self)

        return CustomRequestHandler(request, client_address, server, request_handler=self._request_handler)



from clochette.domain.entity.CalendarID import CalendarID
from clochette.domain.entity.configuration.CalendarConfiguration import CalendarConfiguration


class AuthenticationConfigurationModel:
    _calendar_id: CalendarID
    _edit_calendar: CalendarConfiguration | None

    def __init__(self):
        self._calendar_id = CalendarID.new()
        self._edit_calendar = None

    @property
    def edit_calendar(self):
        return self._edit_calendar

    @edit_calendar.setter
    def edit_calendar(self, calendar: CalendarConfiguration | None):
        self._edit_calendar = calendar
        if calendar is None:
            self._calendar_id = CalendarID.new()

    @property
    def calendar_id(self) -> CalendarID:
        if self._edit_calendar is None:
            return self._calendar_id
        else:
            return self._edit_calendar.id
